import os
import wx
import wx.lib.agw.gradientbutton as GB
import wx.lib.agw.aquabutton as AB
import wx.lib.buttons as buttons


class BCIGui(wx.Frame):
    """BCIGui."""

    def __init__(self, title: str, size: tuple,
                 background_color: str='blue', parent: wx.Frame=None):
        """Init."""
        super(BCIGui, self).__init__(
            parent, title=title, size=size, style=wx.SYSTEM_MENU | wx.CAPTION | wx.CLOSE_BOX)

        self.panel = wx.Panel(self)
        self.SetBackgroundColour(background_color)

        self.buttons = []
        self.input_text = []
        self.static_text = []
        self.images = []
        self.comboboxes = []

    def show_gui(self):
        """Show GUI."""
        self.Show(True)

    def close_gui(self):
        pass

    def add_button(self, message: str, position: tuple, size: tuple,
                   button_type: str=None, color: str=None,
                   action: str='default', id=-1) -> None:
        """Add Button."""
        # Button Type
        if button_type == 'gradient_button':
            btn = GB.GradientButton(
                self.panel, label=message, pos=position, size=size)
        elif button_type == 'aqua_button':
            btn = AB.AquaButton(
                self.panel, label=message, pos=position, size=size)
        else:
            btn = buttons.GenButton(
                self.panel, label=message, pos=position, size=size, id=id)

            # You can really only set colors with GenButtons as the others
            #  use native widgets!
            if color:
                btn.SetBackgroundColour(color)

        # Attach Custom Actions
        self.bind_action(action, btn)

        self.buttons.append(btn)

    def bind_action(self, action: str, btn) -> None:
        """Default action for buttons or comboboxes that do not have an action
        assigned"""
        if action == 'default':
            if isinstance(btn, wx.Button):
                self.Bind(wx.EVT_BUTTON, self.on_clicked, btn)
            elif isinstance(btn, wx.ComboBox):
                self.Bind(wx.EVT_COMBOBOX_DROPDOWN, self.on_dropdown, btn)

    def add_text_input(self, position: tuple, size: tuple) -> None:
        """Add Text Input."""
        input_text = wx.TextCtrl(self.panel, pos=position, size=size)
        self.input_text.append(input_text)

    def add_static_text(self, text: str, position: str,
                        color: str, size: int,
                        font_family: wx.Font=wx.FONTFAMILY_SWISS) -> None:
        """Add Text."""

        static_text = wx.StaticText(
            self.panel, pos=position,
            label=text)
        static_text.SetForegroundColour(color)
        font = wx.Font(size, font_family, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_LIGHT)
        static_text.SetFont(font)

        self.static_text.append(static_text)

    def add_combobox(self, position: tuple, size: tuple, action: str='default') -> None:
        """Add combobox."""
        combobox = wx.ComboBox(self.panel, pos=position, size=size)
        self.comboboxes.append(combobox)
        self.bind_action(action, combobox)

    def add_image(self, path: str, position: tuple, size: int) -> None:
        """Add Image."""
        if os.path.isfile(path):
            img = wx.Image(path, wx.BITMAP_TYPE_ANY)
            # scale the image, preserving the aspect ratio
            width = img.GetWidth()
            height = img.GetHeight()

            if width > height:
                new_width = size
                new_height = size * height / width
            else:
                new_height = size
                new_width = size * width / height

            img = img.Scale(new_width, new_height)
            img = img.ConvertToBitmap()
            bmp = wx.StaticBitmap(self.panel, pos=position, bitmap=img)

            self.images.append(bmp)

        else:
            print('INVALID PATH')

    def on_clicked(self, event):
        """on_clicked.

        Default event to bind to buttons
        """
        btn = event.GetEventObject().GetLabel()
        print(f'pressed {btn}')

    def on_dropdown(self, event):
        """on_dropdown.

        Default event to bind to comboboxes
        """
        print('dropdown selected')


if __name__ == '__main__':
    app = wx.App(False)
    gui = BCIGui(title="BCIGui", size=(650, 650), background_color='black')
    gui.show_gui()
    app.MainLoop()



import numpy as np

def downsample(data: np.array, factor: int=2):
    """Downsamples the data to the given factor

    Parameters:
    -----------
        data : np.array
        factor: amount to downsample
    Returns:
        downsampled data
    """
    return data[:, ::factor]



import errno
import os
from time import localtime, strftime
from shutil import copy2
import json


def init_save_data_structure(data_save_path,
                             user_information,
                             parameters_used,
                             mode=None,
                             experiment_type=None):
    """
    Initialize Save Data Strucutre.

        data_save_path[str]: string of path to save our data in
        user_information[str]: string of user name / related information
        parameters_used[str]: a path to parameters file for the experiment

    """

    # make an experiment folder : note datetime is in utc
    save_folder_name = data_save_path + user_information
    if mode and experiment_type:
        save_directory = save_folder_name + '/' + \
            user_information + '_' + str(mode) + '_' + str(experiment_type) + '_' + strftime(
                '%a_%d_%b_%Y_%Hhr%Mmin%Ssec_%z', localtime())
    else:
        save_directory = save_folder_name + '/' + \
            user_information + '_' + strftime(
                '%a_%d_%b_%Y_%Hhr%Mmin%Ssec_%z', localtime())
    helper_folder_name = save_directory + '/helpers/'

    try:
        # make a directory to save data to
        os.makedirs(save_folder_name)
        os.makedirs(save_directory)
        os.makedirs(helper_folder_name)
        os.makedirs(os.path.join(save_directory, 'logs'), exist_ok=True)

    except OSError as error:
        # If the error is anything other than file existing, raise an error
        if error.errno != errno.EEXIST:
            raise error

        # since this is only called on init, we can make another folder run
        os.makedirs(save_directory)
        os.makedirs(helper_folder_name)
        os.makedirs(os.path.join(save_directory, 'logs'), exist_ok=True)

    try:
        # Go to folder helpers and list files within it
        src_files = os.listdir('bcipy/helpers/')

        # Loop through files in helpers and copy important ones over
        for file_name in src_files:

            # get the full name
            full_file_name = os.path.join('bcipy/helpers/', file_name)

            # Check that constructed file is a real file and ends in .py
            if (os.path.isfile(full_file_name)) and '.py' in file_name:
                # Copy over!
                copy2(full_file_name, helper_folder_name)

        # Check that parameters file given is a real file
        if (os.path.isfile(parameters_used)):
            # Copy over parameters file
            copy2(parameters_used, save_directory)
        else:
            raise Exception('Parameter File Not Found!')

    except IOError as error:
        raise error

    return save_directory


def _save_session_related_data(file, session_dictionary):
    """
    Save Session Related Data.

    Parameters
    ----------
        file[str]: string of path to save our data in
        session_dictionary[dict]: dictionary of session data. It will appear
            as follows:
                {{ "epochs": {
                        "1": {
                          "0": {
                            "copy_phrase": "COPY_PHRASE",
                            "current_text": "COPY_",
                            "eeg_len": 22,
                            "next_display_state": "COPY_",
                            "stimuli": [["+", "_", "G", "L", "B"]],
                            "target_info": [
                              "nontarget", ... ,
                            ],
                            "timing_sti": [[1, 0.2, 0.2, 0.2, 0.2]],
                            "triggers": [[ "+", 0.0], ["_", 0.9922] ..],
                            },
                        ... ,
                        "7": {
                            ... ,
                  },
                  "paradigm": "RSVP",
                  "session": "data/demo_user/demo_user",
                  "session_type": "Copy Phrase",
                  "total_time_spent": 83.24798703193665
                }}
    Returns
    -------
        file, session data file (json file)

    """
    # Try opening as json, if not able to use open() to create first
    try:
        file = json.load(file, 'wt')
    except:
        file = open(file, 'wt')

    # Use the file to dump data to
    try:
        json.dump(session_dictionary, file, indent=2)
    except Exception as e:
        raise e

    return file



import numpy as np
from sklearn.decomposition import PCA


class ChannelWisePrincipalComponentAnalysis:
    """ Channel wise PCA application. Creates PCA objects respective to the
        number of channels.
        Attr:
            n_components(int): Number of components in PCA
            copy(bool): Saves the matrix if  True and updates on each fit()
            whiten(bool): Whitens the PCA matrix to form a tight frame
            svd_solver(string): SV Decomposition solver method
            tol=var_tol(float): Unfortunately I re-implemented it
                Tolerance to the singular values of the matrix
            random_state(seed): Random state seed
            num_ch(int): number of channels in expected data
            var_tol(float): Tolerance to the variance
         """

    def __init__(self, n_components=None, var_tol=0, copy=True, whiten=False,
                 svd_solver='auto', tol=0.0, iterated_power='auto',
                 random_state=None, num_ch=1):

        self.list_pca = []
        for idx in range(num_ch):
            self.list_pca.append(PCA(n_components=n_components, copy=copy,
                                     whiten=whiten,
                                     svd_solver=svd_solver, tol=tol,
                                     iterated_power=iterated_power,
                                     random_state=random_state))
        self.var_tol = var_tol
        self.num_ch = num_ch

    def fit(self, x, y=None, var_tol=None):
        """ Inherits PCA fit() function from scikit-learn. Fits the
            transformation matrix wrt. tolerance to each PCA.
            Args:
                x(ndarray[float]): C x N x k data array
                y(ndarray[int]): N x k observation (class) array
                    N is number of samples k is dimensionality of features
                    C is number of channels
                var_tol(float): Threshold to remove lower variance dims.
                """

        if var_tol:
            self.var_tol = var_tol

        for i in range(self.num_ch):
            self.list_pca[i].fit(x[i, :, :], y)
            max_sv = self.list_pca[i].singular_values_[0]
            self.list_pca[i].n_components = np.sum(self.list_pca[i].singular_values_ >= max_sv * self.var_tol)
            try:
                self.list_pca[i].fit(x[i, :, :], y)
            except Exception as e:
                raise e

    def transform(self, x, y=None):
        f_vector = []
        for i in range(self.num_ch):
            # TODO: Observe that scikit learn PCA does not accept y
            f_vector.append(self.list_pca[i].transform(x[i, :, :]))

        return np.concatenate(f_vector, axis=1)

    def fit_transform(self, x, y=None, var_tol=None):
        """ Fits parameters wrt. the input matrix and outputs corresponding
            reduced form feature vector.
            Args:
                x(ndarray[float]): C x N x k data array
                y(ndarray[int]): N x k observation (class) array
                    N is number of samples k is dimensionality of features
                    C is number of channels
                var_tol(float): Threshold to remove lower variance dims.
            Return:
                y(ndarray(float)): N x ( sum_i (C x k')) data array
                    where k' is the new dimension for each PCA
                """

        self.fit(x, var_tol=var_tol)
        return self.transform(x)


class DummyDimReduction:
    """ Just concatenation without any PCA
        Attr:
        none(None): nothing
         """

    def __init__(self):
        self.none = None

    def fit(self, x, y=None, var_tol=None):
        self.none = None

    def transform(self, x, y=None):
        self.none = None
        num_ch = x.shape[0]
        f_vector = []
        for i in range(num_ch):
            f_vector.append(x[i, :, :])

        return np.concatenate(f_vector, axis=1)

    def fit_transform(self, x, y=None, var_tol=None):
        arg = self.transform(x, y)
        return arg



"""Tests for the DSI headset Protocol."""
import unittest
from bcipy.acquisition.protocols.dsi import dsi
from bcipy.acquisition.protocols.dsi import dsi_protocol


class TestDsiProtocol(unittest.TestCase):
    """Tests for DsiProtocol"""

    def test_protocol_init_messages(self):
        """Should have the channels and the sample_rate."""
        protocol = dsi_protocol.DsiProtocol()
        channel_msg = protocol.init_messages[0]
        parsed1 = dsi.packet.parse(channel_msg)
        self.assertEqual(parsed1.type, 'EVENT')
        self.assertEqual(parsed1.event_code, 'SENSOR_MAP')
        self.assertEqual(parsed1.message, ','.join(dsi.DEFAULT_CHANNELS))

        parsed2 = dsi.packet.parse(protocol.init_messages[1])
        self.assertEqual(parsed2.type, 'EVENT')
        self.assertEqual(parsed2.event_code, 'DATA_RATE')
        self.assertEqual(parsed2.message, u',300')

    def test_encoder(self):
        """It should encode array data that can be subsequently decoded."""

        data = [float(i) for i in range(25)]
        encoder = dsi_protocol.Encoder()
        encoded = encoder.encode(data)
        parsed = dsi.packet.parse(encoded)
        self.assertEqual(parsed.type, 'EEG_DATA')
        self.assertEqual(parsed.sensor_data, data)



import logging


class Task(object):
    """Task.

    Base class for BciPy tasks.
    """

    def __init__(self):
        super(Task, self).__init__()
        self.logger = logging.getLogger(__name__)

    def configure(self):
        pass

    def execute(self):
        raise NotImplementedError()

    def name(self):
        raise NotImplementedError()



import unittest
import numpy as np

from mockito import any, mock, when, unstub

from bcipy.helpers.task import alphabet, trial_reshaper
from bcipy.helpers.load import load_json_parameters


class TestAlphabet(unittest.TestCase):

    def test_alphabet_text(self):
        parameters_used = 'bcipy/parameters/parameters.json'
        parameters = load_json_parameters(parameters_used, value_cast=True)

        parameters['is_txt_stim'] = True

        alp = alphabet(parameters)

        self.assertEqual(
            alp,
            ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',
             'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
             'Y', 'Z', '<',
             '_'])

    def test_alphabet_images(self):
        parameters_used = 'bcipy/parameters/parameters.json'
        parameters = load_json_parameters(parameters_used, value_cast=True)

        parameters['is_txt_stim'] = False
        parameters['path_to_presentation_images'] = ('bcipy/static/images/'
                                                     'rsvp_images/')

        alp = alphabet(parameters)

        self.assertNotEqual(
            alp,
            ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',
             'M', 'N', 'O', 'P', 'R', 'S', 'T', 'U', 'V', 'Y', 'Z', '<',
             '_'])


class TestTrialReshaper(unittest.TestCase):

    def setUp(self):
        self.target_info = ['first_pres_target',
                            'target', 'nontarget', 'nontarget']
        self.timing_info = [1.001, 1.2001, 1.4001, 1.6001]
        # make some fake eeg data
        self.channel_number = 21
        tmp_inp = np.array([range(4000)] * self.channel_number)
        # Add some channel info
        tmp_inp[:, 0] = np.transpose(np.arange(1, 22, 1))
        self.eeg = tmp_inp
        self.channel_map = [1] * self.channel_number

    def tearDown(self):
        unstub()

    def test_trial_reshaper_calibration(self):
        (reshaped_trials, labels,
         num_of_seq, trials_per_sequence) = trial_reshaper(
            trial_target_info=self.target_info,
            timing_info=self.timing_info, eeg_data=self.eeg,
            fs=256, k=2, mode='calibration', channel_map=self.channel_map)

        self.assertTrue(
            len(reshaped_trials) == self.channel_number,
            f'len is {len(reshaped_trials)} not {self.channel_number}')
        self.assertEqual(len(reshaped_trials[0]), 3)
        self.assertEqual(num_of_seq, 1)
        self.assertEqual(trials_per_sequence, 3)

    def test_trial_reshaper_copy_phrase(self):
        pass


if __name__ == '__main__':
    unittest.main()



import logging
import sys
import math
from typing import List
from collections import defaultdict
from bcipy.helpers.task import alphabet, SPACE_CHAR
from bcipy.language_model import lm_server
from bcipy.language_model.lm_server import LmServerConfig
from bcipy.helpers.system_utils import dot
log = logging.getLogger(__name__)
sys.path.append('.')
ALPHABET = alphabet()
LM_SPACE = '#'


class LangModel:
    DEFAULT_CONFIG = LmServerConfig(
        image="lmimage:version2.0",
        port=5000,
        docker_port=5000,
        volumes={dot(__file__, 'fst', 'brown_closure.n5.kn.fst'):
                 "/opt/lm/brown_closure.n5.kn.fst"})

    def __init__(self, server_config: LmServerConfig = DEFAULT_CONFIG,
                 logfile: str = "log"):
        """
        Initiate the langModel class and starts the corresponding docker
        server for the given type.

        Input:
          lmtype - language model type
          logfile - a valid filename to function as a logger
        """
        self.server_config = server_config
        self.priors = defaultdict(list)
        log.setLevel(logging.INFO)
        log.addHandler(logging.FileHandler(logfile))
        lm_server.start(self.server_config)

    def init(self, nbest: int = 1):
        """
        Initialize the language model (on the server side)
        Input:
            nbest - top N symbols from evidence
        """
        lm_server.post_json_request(
            self.server_config, 'init', data={'nbest': nbest})

    def cleanup(self):
        """Stop the docker server"""
        lm_server.stop(self.server_config)

    def reset(self):
        """
        Clean observations of the language model use reset
        """
        lm_server.post_json_request(self.server_config, 'reset')
        self.priors = defaultdict(list)
        log.info("\ncleaning history\n")

    def state_update(self, evidence: List, return_mode: str = 'letter'):
        """
        Provide a prior distribution of the language model
        in return to the system's decision regarding the
        last observation
        Both lm types allow providing more the one timestep
        input. Pay attention to the data struct expected.

        Input:
            decision - a symbol or a string of symbols in encapsulated in a
            list
            the numbers are assumed to be in the log probability domain
            return_mode - 'letter' or 'word' (available
                          for oclm) strings
        Output:
            priors - a json dictionary with Normalized priors
                     in the Negative Log probability domain.
        """
        assert return_mode == 'letter', "PRELM only allows letter output"
        # assert the input contains a valid symbol
        decision = evidence  # in prelm the we treat it as a decision
        for symbol in decision:
            assert symbol in ALPHABET or ' ', \
                "%r contains invalid symbol" % decision
        clean_evidence = []
        for symbol in decision:
            if symbol == SPACE_CHAR:
                symbol = LM_SPACE
            clean_evidence.append(symbol.lower())

        output = lm_server.post_json_request(self.server_config,
                                             'state_update',
                                             {'evidence': clean_evidence,
                                              'return_mode': return_mode})

        return self.__return_priors(output, return_mode)

    def _logger(self):
        """
        Log the priors given the recent decision
        """
        # print a json dict of the priors
        log.info('\nThe priors are:\n')
        for k in self.priors.keys():
            priors = self.priors[k]
            log.info('\nThe priors for {0} type are:\n'.format(k))
            for (symbol, pr) in priors:
                log.info('{0} {1:.4f}'.format(symbol, pr))

    def recent_priors(self, return_mode='letter'):
        """
        Display the priors given the recent decision
        """
        assert return_mode == 'letter', "PRELM only allows letter output"
        if not bool(self.priors[return_mode]):
            output = lm_server.post_json_request(self.server_config,
                                                 'recent_priors',
                                                 {'return_mode': return_mode})
            return self.__return_priors(output, return_mode)
        else:
            return self.priors

    def __return_priors(self, output, return_mode):
        """
        A helper function to provide the desired output 
        depending on the return_mode.
        """

        self.priors = defaultdict(list)
        self.priors['letter'] = [
            (letter.upper(), prob)
            if letter != LM_SPACE  # hard coded (to deal with in future)
            else ("_", prob)
            for (letter, prob) in output['letter']]
        return self.priors



"""DAQ Item Processors"""

import csv
import sys
from bcipy.acquisition.device_info import DeviceInfo


class Processor():
    """Abstract base class for an object that can be used to process data
    acquisition data.
    """

    def __init__(self):
        super(Processor, self).__init__()
        self._device_info = None

    def set_device_info(self, device_info: DeviceInfo):
        """
        Sets the device info, which may be used by the process method.

        Parameters
        ----------
            device_info : DeviceInfo
                Metadata with the parameters used to collect data.
        """
        self._device_info = device_info

    def _check_device_info(self):
        """
        Checks that the sample_metadata has been set.

        Exceptions
        ----------
            throws AssertionError unless set.
        """
        assert self._device_info is not None, \
            "device_info is not set; Initialize with set_device_info."

    def __enter__(self):
        # Device info should be set before using as a context.
        self._check_device_info()
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        pass

    def process(self, record, timestamp=None):
        """Process the given record.

        Parameters
        ----------
            record: list<float>
                A single record returned by from reading a data packet with a
                float value for each channel.
            timestamp: float, optional
                time at which the record was obtained
        """
        raise NotImplementedError('Subclass must define the process method')


class NullProcessor(Processor):
    """A DAQ item Processor that doesn't do anything."""

    def process(self, record, timestamp=None):
        pass


class FileWriter(Processor):
    """A DAQ item Processor that writes items to a file.

    Parameters
    ----------
        filename : str
            Filename to write to.
    """

    def __init__(self, filename):
        super(FileWriter, self).__init__()
        self._filename = filename
        self._file = None
        self._writer = None

    # @override ; context manager
    def __enter__(self):
        self._check_device_info()

        # For python 2, writer needs the 'wb' option in order to work on
        # Windows. If using #Python3 'w' is needed.
        if sys.version_info >= (3, 0, 0):
            self._file = open(self._filename, 'w', newline='')
        else:
            self._file = open(self._filename, 'wb')

        self._writer = csv.writer(self._file, delimiter=',')
        self._writer.writerow(['daq_type', self._device_info.name])
        self._writer.writerow(['sample_rate', self._device_info.fs])
        self._writer.writerow(['timestamp'] + self._device_info.channels)
        return self

    # @override ; context manager
    def __exit__(self, _exc_type, _exc_value, _traceback):
        self._file.close()

    def process(self, record, timestamp=None):
        if self._writer:
            self._writer.writerow([timestamp] + record)


class LslProcessor(Processor):
    """A DAQ item processor that writes to an LSL data stream."""

    def __init__(self):
        super(LslProcessor, self).__init__()
        self._outlet = None

    # @override
    def set_device_info(self, device_info):
        import pylsl
        import uuid

        super(LslProcessor, self).set_device_info(device_info)
        channels = self._device_info.channels
        info = pylsl.StreamInfo(self._device_info.name, 'EEG',
                                len(channels),
                                self._device_info.fs,
                                'float32', str(uuid.uuid4()))
        meta_channels = info.desc().append_child('channels')
        for channel in channels:
            meta_channels.append_child('channel') \
                .append_child_value('label', str(channel)) \
                .append_child_value('unit', 'microvolts') \
                .append_child_value('type', 'EEG')

        self._outlet = pylsl.StreamOutlet(info)

    def process(self, record, timestamp=None):
        if self._outlet:
            self._outlet.push_sample(record, timestamp)


class DispatchProcessor(Processor):
    """Processor that delegates to one or more other processors. Processors
    may be passed in through the constructor or after creating an instance and
    calling it's add method."""

    def __init__(self, *args):
        super(DispatchProcessor, self).__init__()
        self.processors = []
        for proc in args:
            self.add(proc)

    # @override
    def set_device_info(self, device_info):
        super(DispatchProcessor, self).set_device_info(device_info)
        for proc in self.processors:
            proc.set_device_info(device_info)

    def add(self, proc: Processor):
        """Add a processor."""
        self.processors.append(proc)
        if self._device_info:
            proc.set_device_info(self._device_info)

    def remove(self, proc: Processor):
        """Remove a processor."""
        self.processors.remove(proc)

   # @override
    def process(self, record, timestamp=None):
        for proc in self.processors:
            proc.process(record, timestamp)

    # @override ; context manager
    def __enter__(self):
        for proc in self.processors:
            proc.__enter__()
        return self

    # @override ; context manager
    def __exit__(self, _exc_type, _exc_value, _traceback):
        for proc in self.processors:
            proc.__exit__(_exc_type, _exc_value, _traceback)



from psychopy import visual
from bcipy.display.rsvp.display import RSVPDisplay
from bcipy.helpers.task import SPACE_CHAR

""" RSVP Tasks are RSVPDisplay objects with different structure. They share
    the tasks and the essential elements and stimuli. However layout, length of
    stimuli list, update procedures and colors are different. Therefore each
    mode should be separated from each other carefully.
    Functions:
        update_task_state: update task information of the module """


class CopyPhraseDisplay(RSVPDisplay):
    """ Copy Phrase display object of RSVP
        Attr:
            static_task(visual_Text_Stimuli): aim string of the copy phrase.
                (Stored in self.text[0])
            information(visual_Text_Stimuli): information text. (Stored in
                self.text[1])
            task(Multiinfo_color_Stimuli): task visualization.
            sti(visual_Text_Stimuli): stimuli text.
    """

    def __init__(
            self,
            window,
            clock,
            experiment_clock,
            marker_writer,
            static_task_text='COPY_PHRASE',
            static_task_color='White',
            info_text='Press Space Bar to Pause',
            info_color='White',
            info_pos=(0, -.9),
            info_height=0.2,
            info_font='Times',
            task_color=['white'] * 4 + ['green'] * 2 + ['red'],
            task_font='Times',
            task_text='COPY_PH',
            task_height=0.1,
            stim_font='Times',
            stim_pos=(-.8, .9),
            stim_height=0.2,
            stim_sequence=['a'] * 10,
            stim_colors=['white'] * 10,
            stim_timing=[1] * 10,
            is_txt_stim=True,
            trigger_type='image',
            space_char=SPACE_CHAR):
        """ Initializes Copy Phrase Task Objects """

        tmp = visual.TextStim(win=window, font=task_font,
                              text=static_task_text)
        static_task_pos = (
            tmp.boundingBox[0] / window.size[0] - 1, 1 - task_height)

        info_color = [static_task_color, info_color]
        info_font = [task_font, info_font]
        info_text = [static_task_text, info_text]
        info_pos = [static_task_pos, info_pos]
        info_height = [task_height, info_height]

        # Adjust task position wrt. static task position. Definition of
        # dummy texts are required. Place the task on bottom
        tmp2 = visual.TextStim(win=window, font=task_font, text=task_text)
        x_task_pos = tmp2.boundingBox[0] / window.size[0] - 1
        task_pos = (x_task_pos, static_task_pos[1] - task_height)

        super(CopyPhraseDisplay, self).__init__(
            window, clock,
            experiment_clock,
            marker_writer,
            task_color=task_color,
            task_font=task_font,
            task_pos=task_pos,
            task_height=task_height,
            task_text=task_text,
            info_color=info_color,
            info_font=info_font,
            info_pos=info_pos,
            info_height=info_height,
            info_text=info_text,
            stim_font=stim_font,
            stim_pos=stim_pos,
            stim_height=stim_height,
            stim_sequence=stim_sequence,
            stim_colors=stim_colors,
            stim_timing=stim_timing,
            is_txt_stim=is_txt_stim,
            trigger_type=trigger_type,
            space_char=space_char)

    def update_task_state(self, text, color_list):
        """ Updates task state of Copy Phrase Task by removing letters or
            appending to the right.
            Args:
                text(string): new text for task state
                color_list(list[string]): list of colors for each """
        # An empty string will cause an error when we attempt to find its
        # bounding box.
        txt = text if len(text) > 0 else ' '
        tmp2 = visual.TextStim(win=self.window, font=self.task.font, text=txt)
        x_task_pos = tmp2.boundingBox[0] / self.window.size[0] - 1
        task_pos = (x_task_pos, self.text[0].pos[1] - self.task.height)

        self.update_task(text=text, color_list=color_list, pos=task_pos)



"""Interface for creating new device drivers."""
from bcipy.acquisition.device_info import DeviceInfo


class Device():
    """Base class for device-specific behavior.

    Parameters
    ----------
        connection_params : dict
            Parameters needed to connect with the given device
        fs : int
            Sample frequency in Hz.
        channels : list
            List of channel names.
    """

    # pylint: disable=invalid-name
    def __init__(self, connection_params, fs, channels):
        self._connection_params = connection_params
        self.fs = fs
        self.channels = channels

    @property
    def name(self):
        """Device name to be written to the output."""
        raise NotImplementedError('Subclass must define a name property')

    @property
    def device_info(self):
        """Information about the acquisition parameters. Should be called after
        acquisition_init for those devices which set this information."""
        device_name = self.name if not callable(self.name) else self.name()
        return DeviceInfo(fs=self.fs, channels=self.channels, name=device_name)

    def connect(self):
        """Connect to the data source."""
        pass

    def acquisition_init(self):
        """Initialization step. Depending on the protocol, this may involve
        reading header information and setting the appropriate instance
        properties or writing to the server to set params (ex. sampling freq).
        """
        pass

    def read_data(self):
        """Read the next sensor data record from the data source.

        Returns
        -------
            list with a float for each channel.
        """
        raise NotImplementedError(
            'Subclass must define the read_sensor_data method')

    def disconnect(self):
        """Optional method to disconnect from the device and do any cleanup"""
        pass



"""The Buffer is used by the DataAcquisitionClient internally to store data
so it can be queried again. The default buffer uses a Sqlite3 database to
store data. By default it writes to a file called buffer.db, but this can be
configured."""
import os
import sqlite3
from collections import deque
import logging

from builtins import range
from bcipy.acquisition.record import Record

log = logging.getLogger(__name__)


class Buffer():
    """Queryable Data Structure for caching data acquisition data.

    Parameters
    ----------
        channels: list
            list of channel names
        chunksize: int; optional
            Max number of records to keep in memory before flushing to the data
            store. There is a performance tradeoff between average insert time
            (more performant with large chunksize) and max insert time (more
             performant with small chunksize).
        archive_name: str; optional
            Name of the data store.

    http://sebastianraschka.com/Articles/2014_sqlite_in_python_tutorial.html
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, channels, chunksize=10000, archive_name='buffer.db'):

        assert channels, "Buffer wasn't given any channels!"
        assert chunksize > 0, "Chunksize for Buffer must be greater than 0!"
        assert archive_name, "An empty archive name will result in " \
            "an in-memory database that cannot be shared across processes."

        # Flush to data store when buf is this size.
        self._chunksize = chunksize
        self._archive_name = archive_name
        self._buf = deque()
        self.start_time = None
        self._archive_deleted = False

        # There is some discussion and disagreement on re-using sqlite
        # connections between threads. If any issues are encountered, consider
        # opening a new connection for each request, which should be fairly
        # infrequent, or use a connection pool, such as the one provided by
        # the SQLAlchemy library.
        #
        # https://bugs.python.org/issue16509 (check_same_thread parameter)
        # https://stackoverflow.com/questions/14511337/efficiency-of-reopening-sqlite-database-after-each-query
        self._conn = sqlite3.connect(archive_name, check_same_thread=False)
        cursor = self._conn.cursor()

        # Create a data table with the correct number of channels (+ timestamp)
        fields = ['timestamp'] + channels
        self.fields = fields
        # pylint: disable=fixme
        # TODO: Field type should be configurable. For example, if a marker
        # is written to the buffer, this should be a string type.
        defs = ','.join([field + ' real' for field in fields])
        cursor.execute('DROP TABLE IF EXISTS data')

        log.debug(defs)
        cursor.execute('CREATE TABLE data (%s)' % defs)
        self._conn.commit()

        # Create SQL INSERT statement
        field_names = ','.join(fields)
        placeholders = ','.join('?' for i in range(len(fields)))
        self._insert_stmt = (f'INSERT INTO data ({field_names}) '
                             f'VALUES ({placeholders})')

    def _new_connection(self):
        """Returns a new database connection."""
        if self._archive_deleted:
            raise Exception("Queries are not allowed after buffer cleanup. "
                            "Database file has been deleted.")
        return sqlite3.connect(self._archive_name)

    def close(self):
        """Close the buffer."""
        self._flush()

    def cleanup(self, delete_archive=True):
        """Close connection and optionally deletes the archive database.
        Data may not be queryable after this method is called."""

        self._conn.close()
        try:
            if delete_archive:
                os.remove(self._archive_name)
                self._archive_deleted = True
        except OSError:
            pass

    def append(self, record):
        """Append the list of readings for the given timestamp.

        Parameters
        ----------
            record: Record (named tuple)
                sensor reading data and timestamp attributes
        """
        if self.start_time is None:
            self.start_time = record.timestamp

        self._buf.append(record)

        if len(self._buf) >= self._chunksize:
            self._flush()

    def _flush(self):
        """Writes data to the datastore and empties the buffer."""

        data = [_adapt_record(self._buf.popleft())
                for i in range(len(self._buf))]

        if data:
            # Performs writes in a single transaction;
            with self._conn as conn:
                conn.executemany(self._insert_stmt, data)

    def __len__(self):
        self._flush()

        cur = self._new_connection().cursor()
        cur.execute("select count(*) from data", ())
        return cur.fetchone()[0]

    def __str__(self):
        return "Buffer<archive_name={}>".format(self._archive_name)

    def all(self):
        """Returns all data in the buffer."""

        return self.query_data(filters=[("timestamp", ">=", self.start_time)])

    def latest(self, limit=1000):
        """Get the n most recent number of items.

        Parameters
        ----------
            limit : int
                maximum number of items to retrieve
        Returns
        -------
            list of tuple(timestamp, data); will be in descending order by
            timestamp.
        """

        return self.query_data(ordering=("timestamp", "desc"),
                               max_results=limit)

    def query(self, start, end=None, field="_rowid_"):
        """Query the buffer by for a slice of data.

        Parameters
        ----------
            start : number, starting value
            end : number, optional; ending value
                if missing returns the rest of the data
            field : str, optional; query field.
        Returns
        -------
            list of Records
        """
        filters = [(field, ">=", start)]
        if end:
            filters.append((field, "<", end))
        return self.query_data(filters=filters)

    def _validate_query_filters(self, filters):
        """Validates query filters."""
        # Guard against sql injection by limiting query inputs.
        valid_ops = ["<", "<=", ">", ">=", "=", "==", "!=", "<>", "IS",
                     "IS NOT", "IN"]
        valid_fields = ["_rowid_"] + self.fields

        if filters:
            for filter_field, filter_op, _value in filters:
                if filter_field not in valid_fields:
                    raise Exception("Invalid SQL data filter field. Must be "
                                    "one of: " + str(valid_fields))
                elif filter_op not in valid_ops:
                    raise Exception("Invalid SQL operator")

    def _validate_query_ordering(self, ordering):
        """Validates column ordering in sql query
        Parameters:
        -----------
            ordering: tuple(fieldname, direction)
        """

        valid_fields = ["_rowid_"] + self.fields
        valid_directions = ["asc", "desc"]
        if ordering:
            order_field, order_direction = ordering
            if order_field not in valid_fields:
                raise Exception("Invalid field for SQL order by.")
            elif order_direction not in valid_directions:
                raise Exception("Invalid order by direction.")

    def _build_query(self, filters, ordering, max_results):
        """Builds the SQL query with the provided filters.

        Parameters:
        -----------
            filters: list(tuple(field, operator, value)), optional
                list of tuples of the field_name, sql operator, and value.
            ordering: tuple(fieldname, direction), optional
                optional tuple indicating sort order
            max_results: int, optional

        Returns
        -------
            tuple(query:str, params:tuple)
        """
        select = "select _rowid_, * from data"

        where = ''
        query_params = ()
        if filters:
            self._validate_query_filters(filters)
            conditions = " AND ".join([" ".join([field, filter_op, "?"])
                                       for field, filter_op, _ in filters])
            query_params = tuple(filter_val for _, _, filter_val in filters)
            where = "where " + conditions

        order = ''
        if ordering:
            self._validate_query_ordering(ordering)
            order_field, order_direction = ordering
            order = ' '.join(["order by", order_field, order_direction])

        limit = "limit " + \
            str(max_results) if isinstance(max_results, int) else ''

        sql_query = ' '.join(filter(None, [select, where, order, limit]))
        return sql_query, query_params

    def query_data(self, filters=None, ordering=None, max_results=None):
        """Query the data with the provided filters.

        Parameters:
        -----------
            filters: list(tuple(field, operator, value)), optional
                list of tuples of the field_name, sql operator, and value.
            ordering: tuple(fieldname, direction), optional
                optional tuple indicating sort order
            max_results: int, optional

        Returns
        -------
            list of Records matching the query.
        """
        query, query_params = self._build_query(filters, ordering, max_results)

        self._flush()
        conn = self._new_connection()
        result = []
        log.debug('sql query: %s', query)
        log.debug('query params: %s', query_params)
        result = conn.execute(
            query, query_params) if query_params else conn.execute(query)

        # Return the data in the format it was provided
        return [_convert_row(r) for r in result]


def _adapt_record(record):
    """Adapt for insertion into database.

    Parameters
    ----------
        record : Record
    """
    return tuple([record.timestamp] + record.data)


def _convert_row(row):
    """Convert from database row to Record.

    Parameters
    ----------
        row : tuple
    """
    return Record(data=list(row[2:]), timestamp=row[1], rownum=row[0])


def _main():
    import argparse
    import timeit
    from bcipy.acquisition.util import mock_data

    parser = argparse.ArgumentParser()
    parser.add_argument('-n', '--n_records', default=100000, type=int,
                        help='number of records to insert; default is 100000')
    parser.add_argument('-s', '--chunk_size', default=10000, type=int,
                        help="default is 10000")
    parser.add_argument('-c', '--channel_count', default=25, type=int,
                        help="default is 25")

    args = parser.parse_args()
    channels = ["ch" + str(c) for c in range(args.channel_count)]

    print(
        (f"Running with {args.n_records} samples of {args.channel_count} ",
         f"channels each and chunksize {args.chunk_size}"))
    buf = Buffer(channels=channels, chunksize=args.chunk_size)

    starttime = timeit.default_timer()
    for record_data in mock_data(args.n_records, args.channel_count):
        timestamp = timeit.default_timer()
        buf.append(Record(record_data, timestamp, None))

    endtime = timeit.default_timer()
    totaltime = endtime - starttime

    print("Total records inserted: " + str(len(buf)))
    print("Total time: " + str(totaltime))
    print("Records per second: " + str(args.n_records / totaltime))

    print("First 5 records")
    print(buf.query(start=0, end=6))

    buf.cleanup()


if __name__ == '__main__':
    _main()



import sys
import os

from typing import Optional
from pathlib import Path
import pkg_resources
import psutil
import pyglet
import importlib
import pkgutil
import logging


def git_dir() -> str:
    """Git Directory.

    Returns the root directory with the .git folder. If this source code
    was not checked out from scm, answers None."""

    # Relative to current file; may need to be modified if method is moved.
    git_root = Path(os.path.abspath(__file__)).parent.parent.parent
    git_meta = Path(os.path.join(git_root, '.git'))
    return os.path.abspath(git_meta) if git_meta.is_dir() else None


def git_hash() -> Optional[str]:
    """Git Hash.
    
    Returns an abbreviated git sha hash if this code is being run from a
    git cloned version of bcipy; otherwise returns an empty string.

    Could also consider making a system call to:
    git describe --tags
    """

    git_path = git_dir()
    if not git_path:
        print('.git path not found')
        return None

    try:
        head_path = Path(os.path.join(git_path, 'HEAD'))
        with open(head_path) as head_file:
            # First line contains a reference to the current branch.
            # ex. ref: refs/heads/branch_name
            ref = head_file.readline()

        ref_val = ref.split(':')[-1].strip()
        ref_path = Path(os.path.join(git_path, ref_val))
        with open(ref_path) as ref_file:
            sha = ref_file.readline()

        # sha hash is 40 characters; use an abbreviated 7-char version which
        # is displayed in github.
        return sha[0:7]
    except Exception as e:
        print(f'Error reading git version: {e}')
        return None


def bcipy_version() -> str:
    """BciPy Version.

    Gets the current bcipy version. If the current instance of bcipy is a
    git repository, appends the current abbreviated sha hash.
    """
    version = pkg_resources.get_distribution('bcipy').version
    sha_hash = git_hash()

    return f'{version} - {sha_hash}' if sha_hash else version


def get_system_info() -> dict:
    """Get System Information.
    """

    # Three lines for getting screen resolution
    platform = pyglet.window.get_platform()
    display = platform.get_default_display()
    screen = display.get_default_screen()

    mem = psutil.virtual_memory()

    return {
        'os': sys.platform,
        'py_version': sys.version,
        'resolution': [screen.width, screen.height],
        'memory': mem.available / 1024 / 1024,
        'bcipy_version': bcipy_version()
    }


def configure_logger(
        save_folder: str,
        log_name='bcipy_session.txt',
        log_level=logging.DEBUG,
        version=None) -> None:
    """Configure Logger.

    Does what it says.
    """
    # create the log file
    logfile = os.path.join(save_folder, 'logs', log_name)

    # configure it
    logging.basicConfig(
        level=log_level,
        format='(%(threadName)-9s) %(message)s',
        filename=logfile)

    print(f'Printing all BciPy logs to: {logfile}')

    if version:
        logging.info(f'Start of Session for BciPy Version: ({version})')


def import_submodules(package, recursive=True):
    """Import Submodules.

    Import all submodules of a module, recursively, including subpackages.
    https://stackoverflow.com/questions/3365740/how-to-import-all-submodules

    Parameters
    ----------
        package : str | package
            name of package or package instance
        recursive : bool, optional

    Returns
    -------
        dict[str, types.ModuleType]
    """
    if type(package) == str or type(package) == unicode:
        package = importlib.import_module(package)
    results = {}
    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + '.' + name
        if name.startswith('test'):
            continue
        results[full_name] = importlib.import_module(full_name)
        if recursive and is_pkg:
            results.update(import_submodules(full_name))
    return results


def dot(relative_to: str, *argv) -> str:
    """
    Given a file location and one or more subdirectory/filename args, returns
    a Path as a str for that file.

    Ex. dot(__file__, 'fst', 'brown_closure.n5.kn.fst')
    """
    working_dir = Path(os.path.dirname(os.path.realpath(relative_to)))
    for subdir in argv:
        # uses the '/' operator in pathlib to construct a new Path.
        working_dir = working_dir / subdir
    return str(working_dir)


def auto_str(cls):
    """Autogenerate a str method to print all variable names and values.
    https://stackoverflow.com/questions/32910096/is-there-a-way-to-auto-generate-a-str-implementation-in-python
    """
    def __str__(self):
        return '%s(%s)' % (type(self).__name__, ', '.join(
            '%s=%s' % item for item in vars(self).items()))

    cls.__str__ = __str__
    return cls



import unittest

from bcipy.helpers.acquisition import analysis_channel_names_by_pos


class TestAcquisitionRelated(unittest.TestCase):
    """Test helper methods in the acquisition-related code"""

    def testChannelNamesByPos(self):
        names = analysis_channel_names_by_pos(['a', 'b', 'c', 'd'],
                                              [0, 1, 0, 1])
        self.assertEqual('b', names[0])
        self.assertEqual('d', names[1])


# -*- coding: utf-8 -*-
"""Defines the DeviceInfo namedtuple."""
from collections import namedtuple

DeviceInfo = namedtuple('DeviceInfo', ['fs', 'channels', 'name'])
DeviceInfo.__doc__ = """Domain object used for storing metadata about the
collection parameters, including device name, sample rate and the
list of channels."""



import unittest

import tempfile
import shutil
import pickle

from bcipy.helpers.load import (
    load_json_parameters,
    load_experimental_data,
    load_signal_model,
    load_txt_data)


class TestLoad(unittest.TestCase):
    """This is Test Case for Loading BCI data."""

    def setUp(self):
        """set up the needed path for load functions."""

        self.parameters = 'bcipy/parameters/parameters.json'
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def test_load_json_parameters_returns_dict(self):
        """Test load parameters returns a Python dict."""

        # call the load parameters function
        parameters = load_json_parameters(self.parameters)

        # assert that load function turned json parameters into a dict
        self.assertTrue(type(parameters), 'dict')

    def test_load_json_parameters_throws_error_on_wrong_path(self):
        """Test load parameters returns error on entering wrong path."""

        # call the load parameters function with incorrect path
        with self.assertRaises(Exception):
            load_json_parameters('/garbage/dir/wont/work')

    def test_load_classifier(self):
        """Test load classifier can load pickled file when given path."""

        # create a pickle file to save a pickled json
        pickle_file = self.temp_dir + "save.p"
        pickle.dump(self.parameters, open(pickle_file, "wb"))

        # Load classifier
        unpickled_parameters = load_signal_model(pickle_file)

        # assert the same data was returned
        self.assertEqual(unpickled_parameters, (self.parameters, pickle_file))

if __name__ == '__main__':
    unittest.main()


import shutil
import time
import unittest

from bcipy.helpers.acquisition import init_eeg_acquisition
from bcipy.helpers.load import load_json_parameters
from bcipy.helpers.save import init_save_data_structure


class TestAcquisition(unittest.TestCase):

    def setUp(self):
        """set up the needed path for load functions."""
        self.parameters_used = 'bcipy/parameters/parameters.json'
        self.parameters = load_json_parameters(self.parameters_used,
                                               value_cast=True)
        self.data_save_path = 'data/'
        self.user_information = 'test_user_001'

        self.save = init_save_data_structure(
            self.data_save_path,
            self.user_information,
            self.parameters_used)

    def tearDown(self):
        # clean up by removing the data folder we used for testing
        shutil.rmtree(self.save)

    def test_default_values(self):
        print("Testing init_eeg_acquisition with default values.")

        self.parameters['acq_device'] = 'DSI'

        client, server = init_eeg_acquisition(
            self.parameters, self.save, server=True)

        client.start_acquisition()
        time.sleep(0.1)
        client.stop_acquisition()
        client.cleanup()
        server.stop()
        server.join()

        self.assertTrue('raw_data.csv' in client._processor._filename)
        self.assertEqual(client.device_info.name, self.parameters['acq_device'])
        self.assertEqual(client.device_info.fs, 300)

    def test_allows_customization(self):
        print("Testing init_eeg_acquisition with custom values.")

        f = 'foo.csv'
        params = self.parameters
        params['raw_data_name'] = f
        params['acq_port'] = 9000
        params['acq_device'] = 'DSI'

        client, server = init_eeg_acquisition(params, self.save, server=True)

        with client:
            time.sleep(0.1)
        client.cleanup()
        server.stop()
        server.join()

        self.assertTrue(f in client._processor._filename)
        self.assertEqual(client.device_info.name, params['acq_device'])
        self.assertEqual(client.device_info.fs, 300)

    def test_can_use_lsl(self):
        print("Testing init_eeg_acquisition with LSL device")

        params = self.parameters
        params['acq_device'] = 'LSL'

        client, server = init_eeg_acquisition(params, self.save, server=True)

        with client:
            time.sleep(0.1)
        client.cleanup()
        server.stop()
        server.join()

        self.assertEqual(client.device_info.name, 'LSL')
        self.assertEqual(client.device_info.fs, 256)


if __name__ == '__main__':
    unittest.main()



"""Helper functions for language model use."""
import math
from typing import Dict, List, Tuple
import numpy as np
from bcipy.language_model.lm_modes import LangModel, LmType


def init_language_model(parameters):
    """
    Init Language Model.

    Function to Initialize remote language model and get an instance of
        LangModel wrapper. Assumes a docker image is already loaded.

    See language_model/demo/ for more information of how it works.

    Parameters
    ----------
        parameters : dict
            configuration details and path locations

    Returns
    -------
        lmodel: instance
            instance of lmodel wrapper with connections to docker server
    """
    lmtype = LmType[parameters.get("lang_model_type", "PRELM")]

    port = int(parameters['lang_model_server_port'])
    lmodel = LangModel(lmtype, logfile="lmwrap.log", port=port)
    lmodel.init()
    return lmodel


def norm_domain(priors: List[Tuple[str, float]]) -> List[Tuple[str, float]]:
    """Convert a list of (symbol, likelihood) values from negative log
    likelihood to the probability domain (between 0 and 1)

    Parameters:
        priors - list of (symbol, likelihood) values.
            assumes that the units are in the negative log likelihood where
            the lowest value is the most likely.
    Returns:
        list of values in the probability domain (between 0 and 1),
            where the highest value is the most likely.
    """
    return [(sym, math.exp(-prob)) for sym, prob in priors]


def sym_appended(symbol_probs: List[Tuple[str, float]],
                 sym_prob: Tuple[str, float]) -> List[Tuple[str, float]]:
    """Returns a new list of probabilities with the addition of a new symbol
    with the given probability for that symbol. Existing values are adjusted
    equally such that the sum of the probabilities in the resulting list sum
    to approx. 1.0. Only adds the symbol if it is not already in the list.

    Used to add the backspace symbol to the LM output.

    Parameters:
    -----------
        symbol_probs - list of symbol, probability pairs
        sym_prob - (symbol, probability) pair to append
    """
    if sym_prob[0] in dict(symbol_probs):
        return symbol_probs

    # Slit out symbols and probabilities into separate lists
    symbols = [prob[0] for prob in symbol_probs]
    probabilities = np.array([prob[1] for prob in symbol_probs])

    # Add new symbol and its probability
    all_probs = np.append(probabilities, sym_prob[1] / (1 - sym_prob[1]))
    all_symbols = symbols + [sym_prob[0]]

    normalized = all_probs / sum(all_probs)

    return list(zip(all_symbols, normalized))

def equally_probable(alphabet: List[str],
                     specified: Dict[str, float] = None) -> List[float]:
    """Returns a list of probabilities which correspond to the provided
    alphabet. Unless overridden by the specified values, all items will
    have the same probability. All probabilities sum to 1.0.

    Parameters:
    ----------
        alphabet - list of symbols; a probability will be generated for each.
        specified - dict of symbol => probability values for which we want to
            override the default probability.
    Returns:
    --------
        list of probabilities (floats)
    """
    n_letters = len(alphabet)
    if not specified:
        return np.full(n_letters, 1 / n_letters)

    # copy specified dict ignoring non-alphabet items
    overrides = {k: specified[k] for k in alphabet if k in specified}

    prob = (1 - sum(overrides.values())) / (n_letters - len(overrides))
    # override specified values
    return [overrides[sym] if sym in overrides else prob for sym in alphabet]



"""Module for creating server (Process) that manages a data buffer."""
import logging
import multiprocessing as mp
from bcipy.acquisition.buffer import Buffer

# Commands to which the server can respond.
MSG_PUT = 'put_data'
MSG_GET_ALL = 'get_all_data'
MSG_QUERY_SLICE = 'query_slice'
MSG_QUERY = 'query_data'
MSG_COUNT = 'get_count'
MSG_EXIT = 'exit'
MSG_STARTED = 'started'

log = logging.getLogger(__name__)


def _loop(mailbox, channels, archive_name):
    """Main server loop. Intended to be a Process target (and private to this
    module). Accepts messages through its mailbox queue, and takes the
    appropriate action based on the command and parameters contained within the
    message.

    Parameters
    ----------
        mailbox : Queue
            Used for inter-process communication.
        channels : list of str
            list of channel names in the underlying data table. Any records
            written to the buffer are expected to have an entry for each
            channel.
        archive_name : str
            sqlite database name
    """
    buf = Buffer(channels=channels, archive_name=archive_name)

    while True:
        # Messages should be tuples with the structure:
        # (sender, (command, params))
        # where sender is either None or a Queue.
        msg = mailbox.get()
        sender, body = msg
        command, params = body
        if command == MSG_EXIT:
            buf.cleanup(delete_archive=params)
            sender.put(('exit', 'ok'))
            break
        elif command == MSG_PUT:
            # params is the record to put
            buf.append(params)
        elif command == MSG_GET_ALL:
            sender.put(buf.all())
        elif command == MSG_COUNT:
            sender.put(len(buf))
        elif command == MSG_QUERY_SLICE:
            row_start, row_end, field = params
            log.debug("Sending query: %s", (row_start, row_end, field))
            sender.put(buf.query(row_start, row_end, field))
        elif command == MSG_QUERY:
            # Generic query
            filters, ordering, max_results = params
            sender.put(buf.query_data(filters, ordering, max_results))
        elif command == MSG_STARTED:
            sender.put(('started', 'ok'))
        else:
            log.debug("Error; message not understood: %s", msg)


def start(channels, archive_name, asynchronous=False):
    """Starts a server Process.

    Parameters
    ----------
        channels : list of str
            list of channel names. Data records are expected to have an entry
            for each channel.
        archive_name : str
            underlying database name
        asynchronous : boolean, optional; default False
            if true, returns immediately; otherwise waits for a response
            from the newly started server.
    Returns
    -------
        Queue used to communicate with this server instance.
    """

    msg_queue = mp.Queue()
    server_process = mp.Process(target=_loop, args=(
        msg_queue, channels, archive_name))
    server_process.start()
    if not asynchronous:
        request = (MSG_STARTED, None)
        _rpc(msg_queue, request, wait_reply=True)

    return msg_queue


def stop(mailbox, delete_archive=True):
    """Stops the process associated with the provided mailbox.

    Parameters
    ----------
        mailbox : Queue
            queue used to communicate with the process.
        delete_archive : boolean, optional
            if true, deletes the archive on exit
    """
    request = (MSG_EXIT, delete_archive)
    return _rpc(mailbox, request)


def _rpc(mailbox, request, win=None, wait_reply=True):
    """Makes a process call and optionally awaits its reply.

    Parameters
    ----------
        mailbox : Queue
            process queue
        request : tuple
            (command, params), where command is a str and params is a tuple.
        win : Window
            window to reload in case Windows changes focus
        wait_reply : boolean, optional
            waits (blocks) until a response is provided from the server process

    Returns
    -------
        Response from the server or None.
    """
    if wait_reply:
        manager = mp.Manager()
        # Refocus on the window in case Windows changes focus
        if win:
            win.winHandle.activate()

        queue = manager.Queue()

        mailbox.put((queue, request))
        # block until we receive something
        result = queue.get()
        return result

    mailbox.put((None, request))
    return None


def append(mailbox, record):
    """Write the record to the given buffer process. Returns immediately and
     does not wait for a response.

    Parameters
    ----------
        mailbox : Queue
            process queue
        record : Record
            data row to write.
    """
    request = (MSG_PUT, record)
    _rpc(mailbox, request, wait_reply=False)


def count(mailbox):
    """Get the count of the total number of entries in the buffer.

    Parameters
    ----------
        mailbox : Queue
            process queue
    Returns
    -------
        int of the buffer length
    """
    request = (MSG_COUNT, None)
    return _rpc(mailbox, request)

# pylint: disable=redefined-outer-name


def get_data(mailbox, start=None, end=None, field='_rowid_', win=None):
    """Query the buffer for a slice of data records.

    Parameters
    ----------
        mailbox : Queue
            process queue
        start : float, optional
            timestamp of data lower bound; if missing, gets all data
        end : float, optional
            timestamp of data upper bound
        win : Window
            window to pass to _rpc for reloading
    Returns
    -------
        list of data rows within the given range.
    """
    if start is None:
        request = (MSG_GET_ALL, None)
    else:
        request = (MSG_QUERY_SLICE, (start, end, field))
    if win:
        return _rpc(mailbox, request, win)

    return _rpc(mailbox, request)


def query(mailbox, filters=None, ordering=None, max_results=None):
    """Query the buffer for data.

    Parameters:
    -----------
        filters: list(tuple(field, operator, value)), optional
            list of tuples of the field_name, sql operator, and value.
        ordering: tuple(fieldname, direction), optional
            optional tuple indicating sort order
        max_results: int, optional

    Returns
    -------
        list of data rows meeting the query criteria.
    """
    filters = filters or []
    request = (MSG_QUERY, (filters, ordering, max_results))
    return _rpc(mailbox, request)


def main():
    """Test script"""
    import numpy as np
    from bcipy.acquisition.record import Record
    import timeit

    n_rows = 1000
    channel_count = 25
    channels = ["ch" + str(c) for c in range(channel_count)]

    pid1 = start(channels, 'buffer1.db')
    pid2 = start(channels, 'buffer2.db')

    starttime = timeit.default_timer()
    for i in range(n_rows):
        data = [np.random.uniform(-1000, 1000) for _ in range(channel_count)]
        if i % 2 == 0:
            append(pid1, Record(data, i, None))
        else:
            append(pid2, Record(data, i, None))

    endtime = timeit.default_timer()
    totaltime = endtime - starttime

    print("Records inserted in buffer 1: {}".format(count(pid1)))
    print("Records inserted in buffer 2: {}".format(count(pid2)))

    print("Total insert time: " + str(totaltime))

    query_n = 5
    data = get_data(pid1, 0, query_n)
    print("Sample records from buffer 1 (query < {}): {}".format(query_n,
                                                                 data))
    stop(pid1)
    stop(pid2)


if __name__ == '__main__':
    import sys
    if sys.version_info >= (3, 0, 0):
        # Only available in Python 3; allows us to test process code as it
        # behaves in Windows environments.
        mp.set_start_method('spawn')
    main()



from sklearn.neighbors.kde import KernelDensity
import numpy as np


class KernelDensityEstimate:
    """ Kernel density estimate implementation using scikit learn
    library. For further reference, please check scikit learn website.
    Attr:
        bandwidth(float): bandwidth of the kernel
        algorithm(string): algorithm type
        kernel(string): element to form the actual fitted pdf.
        metric(string): distance metric used by algorithm to insert kernels
            onto samples in the given domain.
        atol(float): absolute tolerance to fit a probability distribution
        rtol(float): relative tolerance
        breadth_first(bool): Flag to use breadth first search
        leaf_size(int): Uses a tree model to fit probability density.
        metric_params(dictionary): Used by tree model

        """

    def __init__(self, bandwidth=1.0, algorithm='auto', kernel='gaussian',
                 metric='euclidean', atol=0, rtol=0, breadth_first=True,
                 leaf_size=40, metric_params=None, num_cls=2):
        self.list_den_est = []
        self.num_cls = num_cls
        for i in range(num_cls):
            self.list_den_est.append(KernelDensity(bandwidth=bandwidth,
                                                   algorithm=algorithm,
                                                   kernel=kernel,
                                                   metric=metric,
                                                   atol=atol,
                                                   rtol=rtol,
                                                   breadth_first=breadth_first,
                                                   leaf_size=leaf_size,
                                                   metric_params=metric_params))

    def fit(self, x, y):
        """ Fits the kernel density estimates base on labelled data.
            Attr:
                x(ndarray[float]): N x 1 data array
                y(ndarray[float]): N x 1 label array
                Where N and c denotes number of samples and classes
                respectively. """

        classes = np.unique(y)

        cls_dep_x = [x[np.where(y == classes[i])[0]] for i in
                     range(self.num_cls)]

        for i in range(self.num_cls):
            # Reshape is required, otherwise there's ambiguity if it's one
            # sample with N dims or N samples with 1 dims

            dat = np.squeeze(cls_dep_x[i])
            dat = np.expand_dims(dat, axis=1)

            self.list_den_est[i].fit(dat)

    def transform(self, x):
        """ Calculates likelihood ods of given data.
            Attr:
                x(ndarray[float]): N x 1 data array
                Where N and c denotes number of samples and classes
            Return:
                 val(ndarray[float]): N x c  log-likelihood array
             respectively. """

        # Calculate likelihoods for each density estimate
        val = []
        for i in range(self.num_cls):
            dat = np.squeeze(x)
            dat = np.expand_dims(dat, axis=1)
            val.append(self.list_den_est[i].score_samples(dat))

        return np.transpose(np.array(val))

    def fit_transform(self, x, y):

        self.fit(x, y)
        return self.transform(x)



# pylint: disable=line-too-long
"""Tests for DSI headset driver-related code"""
import unittest
import construct
import pytest
from bcipy.acquisition.protocols.dsi import dsi

# Expected values were generated from existing construct v.2.5.5 structs in the
# existing rsvpkeyboard daq module.


class TestDsi(unittest.TestCase):
    """Tests for binary protocol"""

    def test_header(self):
        """Test encoding/decoding the header info"""
        expected = b'@ABCD\x00\x00\x80\x00\x00\x00\x0c'
        params = dict(type='NULL', payload_length=128, number=12)
        result = dsi.header.build(params)
        self.assertEqual(result, expected)

        parsed = dsi.header.parse(expected)
        for k in params:
            self.assertEqual(parsed[k], params[k])

    def test_none_packet(self):
        """Test null packet"""
        expected = b'@ABCD\x00\x00\x10\x00\x00\x00\r\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f !"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmn'  # noqa
        params = dict(type='NULL', payload_length=16, number=13,
                      none=[i for i in range(111)])
        result = dsi.packet.build(params)
        self.assertEqual(result, expected)

        parsed = dsi.packet.parse(expected)
        for k in params:
            self.assertEqual(parsed[k], params[k])

    def test_event_packet(self):
        """Test encoding/decoding event data"""
        expected = b'@ABCD\x05\x00\x08\x00\x00\x00\r\x00\x00\x00\x02\x00\x00\x00!\x00\x00\x00\x0bhello world'  # noqa
        params = dict(type='EVENT', payload_length=8, number=13,
                      event_code='DATA_START', sending_node=33,
                      message_length=11, message='hello world')

        result = dsi.packet.build(params)
        self.assertEqual(result, expected)

        parsed = dsi.packet.parse(expected)
        for key in params:
            if key != 'message_length':
                self.assertEqual(parsed[key], params[key])

    def test_event_sensor_map(self):
        """Test encoding/decoding list of channels"""
        channels = u'P3,C3,F3,Fz,F4,C4,P4,Cz,CM,A1,Fp1,Fp2,T3,T5,O1,O2,X3,X2,F7,F8,X1,A2,T6,T4,TRG'.encode(  # noqa
            'ascii', 'ignore')
        event_code_bytes, sending_node_bytes, msg_len_bytes, msg_bytes = (
            4, 4, 4, len(channels) * 4)
        payload_length = sum([event_code_bytes, sending_node_bytes,
                              msg_len_bytes, msg_bytes])
        params = dict(type='EVENT', payload_length=payload_length, number=13,
                      event_code='SENSOR_MAP', sending_node=33,
                      message_length=len(channels), message=channels)

        expected = b'@ABCD\x05\x01@\x00\x00\x00\r\x00\x00\x00\t\x00\x00\x00!\x00\x00\x00MP3,C3,F3,Fz,F4,C4,P4,Cz,CM,A1,Fp1,Fp2,T3,T5,O1,O2,X3,X2,F7,F8,X1,A2,T6,T4,TRG'  # noqa

        result = dsi.packet.build(params)
        self.assertEqual(result, expected)

        parsed = dsi.packet.parse(expected)
        channel_names = parsed.message.split(",")
        self.assertEqual(len(channel_names), 25)

    def test_eeg_packet(self):
        """Test encoding/decoding EEG_DATA packet"""
        expected = b'@ABCD\x01\x00\x13\x00\x00\x00\rN\x05Y\xe9\x00123456C\xf2ff\xc3\xfb\x19\x9a'  # noqa

        params = dict(type='EEG_DATA', payload_length=19, number=13,
                      timestamp=559315520.0, data_counter=0,
                      ADC_status=b"123456", sensor_data=[484.799987793,
                                                         -502.200012207])
        result = dsi.packet.build(params)
        self.assertEqual(result, expected)

        parsed = dsi.packet.parse(expected)
        for key in params:
            if key != 'sensor_data':
                self.assertEqual(parsed[key], params[key])
        self.assertEqual(("%.9f" % parsed['sensor_data'][0]), '484.799987793')
        self.assertEqual(("%.9f" % parsed['sensor_data'][1]), '-502.200012207')

        # Payload size that exceeds sensor_data points should throw an error
        with pytest.raises(construct.core.RangeError):
            params['payload_length'] = 24
            dsi.packet.build(params)




class ConnectionErr(Exception):
    def __init__(self, host, port):
        self.dErrhost = host
        self.dErrport = port
        Exception.__init__(
            self, "Connection was not established with host {0} and port {1}".format(
                self.dErrhost, self.dErrport))


class StatusCodeError(Exception):
    def __init__(self, status_code):
        self.dErrArgs = status_code
        Exception.__init__(
            self,
            "Connection was not established with and has status code {0}".format(
                self.dErrArgs))


class DockerDownError(Exception):
    def __init__(self):
        Exception.__init__(self, "Check that Docker is up and running")


class EvidenceDataStructError(Exception):
    def __init__(self):
        Exception.__init__(
            self, "The evidence data structure is incorrect. It should be of [[(s,p),(s,p)]] or [[(s,p),(s,p)],[(s,p),(s,p)]]{s - a valid string, p - float}")


class NBestError(Exception):
    def __init__(self, nbest):
        self.dErrArgs = nbest
        Exception.__init__(
            self,
            "Invalid nbest input of {0}. must be an integer".format(
                self.dErrArgs))


class NBestHighValue(Warning):
    def __init__(self, nbest):
        self.nbest = nbest
        Warning.__init__(
            self, "nbest of {0} is higher than 4 and may not perform optimally due to memory limitations".format(nbest))



"""Data Acquisition module. This module streams data from the EEG hardware,
persists it in csv files, and makes it available to other systems."""
import sys
from os.path import dirname
sys.path.append(dirname(__file__))
sys.path.append('.')
sys.path.append('..')



# pylint: disable=invalid-name
"""Protocol with information for generating and serving mock DSI data."""
import timeit

from bcipy.acquisition.protocols.dsi import dsi
from bcipy.acquisition.protocols import protocol


def DsiProtocol(fs=dsi.DEFAULT_FS, channels=None):
    """Protocol for mocking DSI data."""
    channels = channels or dsi.DEFAULT_CHANNELS
    return protocol.Protocol(encoder=Encoder(),
                             init_messages=_init_messages(fs, channels),
                             fs=fs,
                             channels=channels)


def _event_packet(code, msg):
    """Construct an event packet with the given event code and message."""

    # calculate payload length
    event_code_bytes, sending_node_bytes, msg_len_bytes, msg_bytes = (
        4, 4, 4, len(msg))

    payload_len = sum(
        [event_code_bytes, sending_node_bytes, msg_len_bytes, msg_bytes])

    params = dict(type='EVENT', payload_length=payload_len, number=13,
                  event_code=code, sending_node=33,
                  message_length=len(msg), message=msg)

    return dsi.packet.build(params)


def _init_messages(fs, channels):
    """Messages sent at the start of the initialization protocol when
    connecting with a client. Sent before any data is sent.

    Parameters
    ----------
        fs : int
            sample frequency
        channels : list
            list of channel names
    """

    s_msg = ','.join(channels).encode('ascii', 'ignore')
    sensor_msg = _event_packet('SENSOR_MAP', s_msg)

    # Comma-delimited list; the frequency is the second param, which is the
    # only one we use; not sure what else is in this list.
    f_msg = (',' + str(fs)).encode('ascii', 'ignore')
    freq_msg = _event_packet('DATA_RATE', f_msg)

    return [sensor_msg, freq_msg]


class Encoder():
    """Encodes sensor data as binary data in the DSI format."""

    def __init__(self):
        super(Encoder, self).__init__()
        self.start_time = timeit.default_timer()

    def encode(self, sensor_data):
        """Builds a binary data packet from the provided sensor data. Timestamp
        is the difference from the given start_time.

        Parameters
        ----------
            sensor_data: list - list of sensor values (float type); len must
                match the channel_count

        Returns
        -------
        Binary data for a single packet.
        """
        # (timestamp/4 bytes, counter/1 byte, status/6 bytes) + 4 bytes for
        # every sensor/float.
        payload_length = (11 + (len(sensor_data) * 4))

        params = dict(type='EEG_DATA',
                      payload_length=payload_length,
                      number=13,  # arbitrary numberr
                      timestamp=(timeit.default_timer() - self.start_time),
                      data_counter=0,  # unused
                      ADC_status=b"123456",  # unused
                      sensor_data=sensor_data)

        return dsi.packet.build(params)



# pylint: disable=fixme,too-many-instance-attributes,too-many-arguments
"""Data Acquisition Client"""
import logging
import multiprocessing
import time

from queue import Empty

from bcipy.acquisition import buffer_server
from bcipy.acquisition.processor import FileWriter
from bcipy.acquisition.record import Record
from bcipy.acquisition.util import StoppableProcess
from bcipy.acquisition.marker_writer import NullMarkerWriter, LslMarkerWriter

log = logging.getLogger(__name__)
DEBUG = False
DEBUG_FREQ = 500
MSG_DEVICE_INFO = "device_info"
MSG_ERROR = "error"
MSG_PROCESSOR_INITIALIZED = "processor_initialized"


class CountClock():
    """Clock that provides timestamp values starting at 1.0; the next value
    is the increment of the previous. Implements the monotonic clock interface.
    """

    def __init__(self):
        super(CountClock, self).__init__()
        self.counter = 0

    def reset(self):
        """Resets the counter"""
        self.counter = 0

    # pylint: disable=invalid-name
    def getTime(self):
        """Gets the current count."""
        self.counter += 1
        return float(self.counter)


class DataAcquisitionClient:
    """Data Acquisition client. The client sets up a separate thread for
    acquisition, writes incoming data to a queue, and processes the data from
    the queue.

    Parameters
    ----------
        device: Device instance
            Object with device-specific implementations for connecting,
            initializing, and reading a packet.
        processor : Processor; optional
            A data Processor that does something with the streaming data (ex.
            writes to a file.)
        buffer_name : str, optional
            Name of the sql database archive; default is buffer.db.
        clock : Clock, optional
            Clock instance used to timestamp each acquisition record
        delete_archive: boolean, optional
            Flag indicating whether to delete the database archive on exit.
            Default is True.
    """

    def __init__(self,
                 device,
                 processor=FileWriter(filename='rawdata.csv'),
                 buffer_name='buffer.db',
                 clock=CountClock(),
                 delete_archive=True):

        self._device = device
        self._processor = processor
        self._buffer_name = buffer_name
        self._clock = clock

        # boolean; set to false to retain the sqlite db.
        self.delete_archive = delete_archive

        self._device_info = None  # set on start.
        self._is_streaming = False

        # Offset in seconds from the start of acquisition to calibration
        # trigger. Calculated once, then cached.
        self._cached_offset = None
        self._record_at_calib = None
        self._max_wait = 0.1  # for process loop

        # Max number of records in queue before it blocks for processing.
        maxsize = 500

        self._process_queue = multiprocessing.JoinableQueue(maxsize=maxsize)
        self.marker_writer = NullMarkerWriter()
        self._acq_process = None
        self._data_processor = None
        self._buf = None

    # @override ; context manager
    def __enter__(self):
        self.start_acquisition()
        return self

    # @override ; context manager
    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.stop_acquisition()

    def start_acquisition(self):
        """Run the initialization code and start the loop to acquire data from
        the server.

        We use multiprocessing to achieve best performance during our sessions.

        Some references:
            Stopping processes and other great multiprocessing examples:
                https://pymotw.com/2/multiprocessing/communication.html
            Windows vs. Unix Process Differences:
                https://docs.python.org/2.7/library/multiprocessing.html#windows

        """

        if not self._is_streaming:
            log.debug("Starting Acquisition")

            msg_queue = multiprocessing.Queue()

            # Initialize the marker streams before the device connection so the
            # device can start listening.
            # TODO: Should this be a property of the device?
            if self._device.name == 'LSL':
                self.marker_writer = LslMarkerWriter()

            # Clock is copied, so reset should happen in the main thread.
            self._clock.reset()

            self._acq_process = AcquisitionProcess(self._device, self._clock,
                                                   self._process_queue,
                                                   msg_queue)
            self._acq_process.start()

            # Block thread until device connects and returns device_info.
            msg_type, msg = msg_queue.get()
            if msg_type == MSG_DEVICE_INFO:
                self._device_info = msg
            elif msg_type == MSG_ERROR:
                raise Exception("Error connecting to device")
            else:
                raise Exception("Message not understood: " + str(msg))

            # Initialize the buffer and processor; this occurs after the
            # device initialization to ensure that any device parameters have
            # been updated as needed.
            self._processor.set_device_info(self._device_info)
            self._buf = buffer_server.start(self._device_info.channels,
                                            self._buffer_name)

            self._data_processor = DataProcessor(
                data_queue=self._process_queue,
                msg_queue = msg_queue,
                processor=self._processor,
                buf=self._buf,
                wait=self._max_wait)
            self._data_processor.start()

            # Block until processor has initialized.
            msg_type, msg = msg_queue.get()
            self._is_streaming = True

    def stop_acquisition(self):
        """Stop acquiring data; perform cleanup."""
        log.debug("Stopping Acquisition Process")

        self._is_streaming = False

        self._acq_process.stop()
        self._acq_process.join()

        log.debug("Stopping Processing Queue")

        # Blocks until all data in the queue is consumed.
        self._process_queue.join()
        self._data_processor.stop()
        self.marker_writer.cleanup()
        self.marker_writer = NullMarkerWriter()

    def get_data(self, start=None, end=None, field='_rowid_', win=None):
        """Queries the buffer by field.

        Parameters
        ----------
            start : number, optional
                start of time slice; units are those of the acquisition clock.
            end : float, optional
                end of time slice; units are those of the acquisition clock.
            field: str, optional
                field on which to query; default value is the row id.
            win : Window
                window to pass to server for reloading
        Returns
        -------
            list of Records
        """

        if self._buf is None:
            return []

        return buffer_server.get_data(self._buf, start, end, field, win)

    def get_data_for_clock(self, calib_time: float, start_time: float,
                           end_time: float):
        """Queries the database, using start and end values relative to a
        clock different than the acquisition clock.

        Parameters
        ----------
            calib_time: float
                experiment_clock time (in seconds) at calibration.
            start_time : float, optional
                start of time slice; units are those of the experiment clock.
            end_time : float, optional
                end of time slice; units are those of the experiment clock.
        Returns
        -------
            list of Records
        """

        sample_rate = self._device_info.fs

        if self._record_at_calib is None:
            rownum_at_calib = 1
        else:
            rownum_at_calib = self._record_at_calib.rownum

        # Calculate number of samples since experiment_clock calibration;
        # multiplying by the fs converts from seconds to samples.
        start_offset = (start_time - calib_time) * sample_rate
        start = rownum_at_calib + start_offset

        end = None
        if end_time:
            end_offset = (end_time - calib_time) * sample_rate
            end = rownum_at_calib + end_offset

        return self.get_data(start=start, end=end, field='_rowid_')

    def get_data_len(self):
        """Efficient way to calculate the amount of data cached."""
        if self._buf is None:
            return 0
        return buffer_server.count(self._buf)

    @property
    def device_info(self):
        """Get the latest device_info."""
        return self._device_info

    @property
    def is_calibrated(self):
        """Returns boolean indicating whether or not acquisition has been
        calibrated (an offset calculated based on a trigger)."""
        return self.offset is not None

    @is_calibrated.setter
    def is_calibrated(self, bool_val):
        """Setter for the is_calibrated property that allows the user to
        override the calculated value and use a 0 offset.

        Parameters
        ----------
            bool_val: boolean
                if True, uses a 0 offset; if False forces the calculation.
        """
        self._cached_offset = 0.0 if bool_val else None

    @property
    def offset(self):
        """Offset in seconds from the start of acquisition to calibration
        trigger.

        Returns
        -------
            float or None if TRG channel is all 0.
        TODO: Consider setting the trigger channel name in the device_info.
        """

        # cached value if previously queried; only needs to be computed once.
        if self._cached_offset is not None:
            return self._cached_offset

        if self._buf is None or self._device_info is None:
            log.debug("Buffer or device has not been initialized")
            return None

        log.debug("Querying database for offset")
        # Assumes that the TRG column is present and used for calibration, and
        # that non-calibration values are all 0.
        rows = buffer_server.query(self._buf,
                                   filters=[("TRG", ">", 0)],
                                   ordering=("timestamp", "asc"),
                                   max_results=1)
        if not rows:
            log.debug("No rows have a TRG value.")
            return None

        log.debug(rows[0])
        # Calculate offset from the number of samples recorded by the time
        # of calibration.
        self._record_at_calib = rows[0]
        self._cached_offset = rows[0].rownum / self._device_info.fs
        log.debug("Cached offset: %s", str(self._cached_offset))
        return self._cached_offset

    @property
    def record_at_calib(self):
        """Data record at the calibration trigger"""
        return self._record_at_calib

    def cleanup(self):
        """Performs cleanup tasks, such as deleting the buffer archive. Note
        that data will be unavailable after calling this method."""
        if self._buf:
            buffer_server.stop(self._buf, delete_archive=self.delete_archive)
            self._buf = None


class AcquisitionProcess(StoppableProcess):
    """Process that continuously reads data from the data source and sends it
    to the data_queue for processing.


    Parameters
    ----------
        device: Device instance
            Object with device-specific implementations for connecting,
            initializing, and reading a packet.
        clock : Clock
            Used to timestamp each record as it is read.
        data_queue : Queue
            Data will be written to the queue as it is acquired.
        msg_queue : Queue
            Used to communicate messages to the main thread.
    """

    def __init__(self, device, clock, data_queue, msg_queue):
        super(AcquisitionProcess, self).__init__()
        self._device = device
        self._clock = clock
        self._data_queue = data_queue
        self._msg_queue = msg_queue

    def run(self):
        """Process startup. Connects to the device and start reading data.
        Since this is done in a separate process from the main thread, any
        errors encountered will be written to the msg_queue.
        """

        try:
            log.debug("Connecting to device")
            self._device.connect()
            self._device.acquisition_init()
        except Exception as error:
            self._msg_queue.put((MSG_ERROR, str(error)))
            raise error

        # Send updated device info to the main thread; this also signals that
        # initialization is complete.
        self._msg_queue.put((MSG_DEVICE_INFO, self._device.device_info))

        log.debug("Starting Acquisition read data loop")
        sample = 0
        data = self._device.read_data()

        # begin continuous acquisition process as long as data received
        while self.running() and data:
            sample += 1
            if DEBUG and sample % DEBUG_FREQ == 0:
                log.debug("Read sample: %s", str(sample))

            self._data_queue.put(Record(data, self._clock.getTime(), sample))
            try:
                # Read data again
                data = self._device.read_data()
            # pylint: disable=broad-except
            except Exception as error:
                log.error("Error reading data from device: %s", str(error))
                data = None
                break
        log.debug("Total samples read: %s", str(sample))
        self._device.disconnect()


class DataProcessor(StoppableProcess):
    """Process that gets data from a queue and persists it to the buffer."""

    def __init__(self, data_queue, msg_queue, processor, buf, wait=0.1):
        super(DataProcessor, self).__init__()
        self._data_queue = data_queue
        self._msg_queue = msg_queue
        self._processor = processor
        self._buf = buf
        self._wait = wait

    def run(self):
        """Reads from the queue of data and performs processing an item at a
        time. Also writes data to buffer."""

        log.debug("Starting Data Processing loop.")
        count = 0
        with self._processor as processor:
            # Signal that initialization is complete.
            self._msg_queue.put((MSG_PROCESSOR_INITIALIZED, True))

            while self.running():
                try:
                    record = self._data_queue.get(True, self._wait)
                    count += 1
                    if DEBUG and count % DEBUG_FREQ == 0:
                        log.debug("Processed sample: %s", str(count))
                    buffer_server.append(self._buf, record)
                    processor.process(record.data, record.timestamp)
                    self._data_queue.task_done()
                except Empty:
                    pass
            log.debug("Total samples processed: %s", str(count))


def main():
    """Test script."""
    import sys
    if sys.version_info >= (3, 0, 0):
        # Only available in Python 3; allows us to test process code as it
        # behaves in Windows environments.
        multiprocessing.set_start_method('spawn')

    import argparse
    import json
    from bcipy.acquisition.protocols import registry

    parser = argparse.ArgumentParser()
    parser.add_argument('-b', '--buffer', default='buffer.db',
                        help='buffer db name')
    parser.add_argument('-f', '--filename', default='rawdata.csv')
    parser.add_argument('-d', '--device', default='DSI',
                        choices=registry.supported_devices.keys())
    parser.add_argument('-c', '--channels', default='',
                        help='comma-delimited list')
    parser.add_argument('-p', '--params', type=json.loads,
                        default={'host': '127.0.0.1', 'port': 9000},
                        help="device connection params; json")
    args = parser.parse_args()

    device_builder = registry.find_device(args.device)

    # Instantiate and start collecting data
    dev = device_builder(connection_params=args.params)
    if args.channels:
        dev.channels = args.channels.split(',')
    daq = DataAcquisitionClient(device=dev,
                                processor=FileWriter(filename=args.filename),
                                buffer_name=args.buffer,
                                delete_archive=True)

    daq.start_acquisition()

    # Get data from buffer
    time.sleep(1)

    print("Number of samples in 1 second: {0}".format(daq.get_data_len()))

    time.sleep(1)

    print("Number of samples in 2 seconds: {0}".format(daq.get_data_len()))

    daq.stop_acquisition()
    daq.cleanup()


if __name__ == "__main__":
    main()



"""Functions for generating mock data to be used for testing/development."""

import logging

from past.builtins import range
from bcipy.signal.generator.generator import gen_random_data

log = logging.getLogger(__name__)


def advance_to_row(filehandle, rownum):
    """Utility function to advance a file cursor to the given row."""
    for _ in range(rownum - 1):
        filehandle.readline()

# pylint: disable=too-few-public-methods


class _DefaultEncoder():
    """Encodes data by returning the raw data."""

    # pylint: disable=no-self-use
    def encode(self, data):
        """Encode the data that will be output by the file_data generator."""
        return data


def random_data(encoder=_DefaultEncoder(), low=-1000, high=1000,
                channel_count=25):
    """Generates random EEG-like data encoded according to the provided encoder.

    Returns
    -------
        packet of data, which decodes into a list of floats in the range low
            to high with channel_count number of items.
    """

    while True:
        sensor_data = gen_random_data(low, high, channel_count)
        yield encoder.encode(sensor_data)


def file_data(filename, header_row=3, encoder=_DefaultEncoder()):
    """Generates data from a source file and encodes it according to the
    provided encoder.

    Parameters
    ----------
        filename: str
            Name of file containing a sample EEG session output. This file will
            be the source of the generated data. File should be a csv file.
        header_row: int, optional
            Row with the header data (channel names); the default is 3,
            assuming the first 2 rows are for metadata.
        encoder : Encoder, optional
            Used to encode the output data.

    """

    with open(filename, 'r') as infile:
        # advance to first data row, since channel names are unused.
        advance_to_row(infile, header_row + 1)

        while True:
            line = infile.readline()
            if not line:
                break
            sensor_data = list(map(data_value, line.split(",")))            
            yield encoder.encode(sensor_data)

def data_value(value: str) -> float:
    """Convert to a float; some trigger values are strings, rather than 
    numbers (ex. indicating the letter); convert these to 1.0."""
    if value:
        try:
            return float(value)
        except ValueError:
            return 1.0
    else:
        # empty string
        return 0.0


import os

from scipy.signal import butter, sosfilt
import logging
import numpy as np

log = logging.getLogger(__name__)


def butter_bandpass(lowcut, highcut, fs, order=5):
        nyq = 0.5 * fs
        low = lowcut / nyq
        high = highcut / nyq
        sos = butter(order, [low, high], analog=False, btype='band', output='sos')
        return sos


def butter_bandpass_filter(data, lowcut, highcut, fs, order=5):
        sos = butter_bandpass(lowcut, highcut, fs, order=order)
        y = sosfilt(sos, data)
        return y


def text_filter(input_seq, filt=None, fs=256, k=2, filter_location=None):
    """
    :param input_seq: Input sequence to be filtered. Expected dimensions are 16xT
    :param filt: Input for using a specific filter. If left empty, according to fs a pre-designed filter is going to be used. Filters are pre-designed for fs = 256,300 or 1024 Hz.
    :param fs: Sampling frequency of the hardware.
    :param k: downsampling order
    :param filter_location: Path to filters.txt, If left empty, filters.txt is assumed to be next to this sig_pro.py

    :return: output sequence that is filtered and downsampled input. Filter delay is compensated. Dimensions are 16xT/k

    256Hz
        - 1.75Hz to 45Hz
        - 60Hz -64dB Gain

    300Hz
        - 1.84Hz to 45Hz
        - 60Hz -84dB Gain

    1024Hz
        - 1.75Hz to 45Hz
        - 60Hz -64dB Gain

    """

    # If filter location is not provided, assume it is next to sig_pro.py file.
    if not filter_location:
        filter_location = os.path.dirname(os.path.abspath(__file__)) + '/resources/filters.txt'

    # Try to open the filters.txt file
    try:
        with open(filter_location, 'r') as text_file:
            dict_of_filters = eval(text_file.readline())
    except Exception as e:
        log.error('filters.txt cannot be found in path that is passed:', filter_location)
        raise e

    # Try to get the required filter from the text file.
    try:
        filt = dict_of_filters[fs]
    except Exception as e:
        log.error('filters.txt does not have a filter with sampling frequency provided.')
        raise e

    # Precision correction
    filt = np.array(filt)
    filt = filt - np.sum(filt) / filt.size

    # Initialize output sequence
    output_seq = [[]]

    # Convolution per channel
    for z in range(len(input_seq)):
        temp = np.convolve(input_seq[z][:], filt)
        # Filter off-set compensation
        temp = temp[int(np.ceil(len(filt) / 2.)) - 1:]
        # Downsampling
        output_seq.append(temp[::k])

    return np.array(output_seq[1:])



"""TCP Server to stream mock EEG data."""

import logging
from queue import Queue, Empty
import select
import socket
import time


from bcipy.acquisition.datastream.producer import Producer
from bcipy.acquisition.util import StoppableThread

log = logging.getLogger(__name__)


class DataServer(StoppableThread):
    """Streams sample EEG-like data via TCP.

    Parameters
    ----------
        protocol : Protocol
            protocol-specific behavior including fs, init_messages, and encoder
        generator : function
            generator function; used to generate the data to be served. A new
            generator will be created for every client connection.
        gen_params : dict, optional
            parameters for the generator function
        host : str, optional
        port : int, optional
    """

    # pylint: disable=too-many-arguments
    def __init__(self, protocol, generator, gen_params=None, host='127.0.0.1',
                 port=9999):
        super(DataServer, self).__init__(name="DataServer")
        self.protocol = protocol
        self.generator = generator
        self.gen_params = gen_params or {}
        self.gen_params['encoder'] = protocol.encoder

        self.host = host
        self.port = port

        self.started = False

        # Putting socket.bind in constructor allows us to handle any errors
        # (such as Address already in use) in the calling context.
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(
            socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_socket.bind((self.host, self.port))

    def run(self):
        """Listen for client connection and initiate handler."""

        self.server_socket.listen(1)

        self.started = True
        log.debug("[*] Accepting connections on %s:%d", self.host,
                      self.port)

        while self.running():

            # Sockets from which we expect to read
            inputs = [self.server_socket]
            outputs = []
            errors = []

            # Polls for readable state
            readable, _w, _e = select.select(inputs, outputs, errors, 0.05)

            if readable:
                client, addr = self.server_socket.accept()
                log.debug("[*] Accepted connection from: %s:%d", addr[0],
                              addr[1])
                self._handle_client(client)
        try:
            self.server_socket.shutdown(2)
            self.server_socket.close()
        # pylint: disable=broad-except
        except Exception:
            pass
        self.server_socket = None
        log.debug("[*] No longer accepting connections")

    def stop(self):
        log.debug("[*] Stopping data server")
        super(DataServer, self).stop()

    def _handle_client(self, client_socket):
        """This currently only handles a single client and blocks on that call.

        If multiple clients need to connect, the server needs to track the
        clients and send the data to each one (each of which would have a queue
        of data to be sent)
        """
        wfile = client_socket.makefile(mode='wb')

        # Send initialization data according to the protocol
        for msg in self.protocol.init_messages:
            wfile.write(msg)

        # This needs to be handled differently if we decide to allow
        # multiple clients. Producer needs to be managed by the class,
        # and the same message should be sent to each client.
        data_queue = Queue()

        # Construct a new generator each time to get consistent results.
        generator = self.generator(**self.gen_params)
        with Producer(data_queue, generator=generator,
                      freq=1 / self.protocol.fs):
            while self.running():
                try:
                    # block if necessary, for up to 5 seconds
                    item = data_queue.get(True, 5)
                except Empty:
                    client_socket.close()
                    break
                try:
                    wfile.write(item)
                except IOError:
                    break
        client_socket.close()
        log.debug("[*] Client disconnected")


def start_socket_server(protocol, host, port, retries=2):
    """Starts a DataServer given the provided port and host information. If
    the port is not available, will automatically try a different port up to
    the given number of times. Returns the server along with the port.

    Parameters
    ----------
        protocol : Protocol for how to generate data.
        host : str ; socket host (ex. '127.0.0.1').
        port : int.
        retries : int; number of times to attempt another port if provided
            port is busy.
    Returns
    -------
        (server, port)
    """
    from bcipy.acquisition.datastream.generator import random_data
    try:
        dataserver = DataServer(protocol=protocol,
                                generator=random_data,
                                gen_params={'channel_count': len(
                                    protocol.channels)},
                                host=host,
                                port=port)

    except IOError as error:
        if retries > 0:
            # try a different port when 'Address already in use'.
            port = port + 1
            log.debug("Address in use: trying port %d", port)
            return start_socket_server(protocol, host, port, retries - 1)
        raise error

    await_start(dataserver)
    return dataserver, port


def await_start(dataserver, max_wait=2):
    """Blocks until server is started. Raises if max_wait is exceeded before
    server is started."""

    dataserver.start()
    wait = 0
    wait_interval = 0.01
    while not dataserver.started:
        time.sleep(wait_interval)
        wait += wait_interval
        if wait >= max_wait:
            dataserver.stop()
            raise Exception("Server couldn't start up in time.")


def _settings(filename):
    """Read the daq settings from the given data file"""

    with open(filename, 'r') as infile:
        daq_type = infile.readline().strip().split(',')[1]
        sample_hz = int(infile.readline().strip().split(',')[1])
        channels = infile.readline().strip().split(',')
        return (daq_type, sample_hz, channels)


def main():
    """Initialize and run the server."""
    import argparse

    from bcipy.acquisition.datastream.generator import file_data, random_data
    from bcipy.acquisition.protocols.registry import protocol_with, \
        default_protocol

    parser = argparse.ArgumentParser()
    parser.add_argument('-H', '--host', default='127.0.0.1')
    parser.add_argument('-p', '--port', type=int, default=8844)
    parser.add_argument('-f', '--filename', default=None,
                        help="file containing data to be streamed; "
                        "if missing, random data will be served.")
    args = parser.parse_args()

    if args.filename:
        daq_type, sample_hz, channels = _settings(args.filename)
        protocol = protocol_with(daq_type, sample_hz, channels)
        generator, params = (file_data, {'filename': args.filename})
    else:
        protocol = default_protocol('DSI')
        channel_count = len(protocol.channels)
        generator, params = (random_data, {'channel_count': channel_count})

    try:
        server = DataServer(protocol=protocol,
                            generator=generator,
                            gen_params=params,
                            host=args.host, port=args.port)
        log.debug("New server created")
        server.start()
        log.debug("Server started")
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("Keyboard Interrupt")
        server.stop()


if __name__ == '__main__':
    # Run this as: python -m acquisition.datastream.server
    main()



# pylint: disable=fixme,invalid-name
"""Used to find a protocol or device by name."""

from bcipy.helpers.system_utils import import_submodules
from bcipy.acquisition.protocols.device import Device
from bcipy.acquisition.protocols.dsi.dsi_protocol import DsiProtocol

# # import all submodules so we can introspect on subclasses.
import_submodules('protocols')


def _key(pyclass):
    """
    >>> from acquisition.protocols.dsi_device import DsiDevice
    >>> _key(DsiDevice)
    u'DSI'
    """
    return pyclass.__module__.split(".")[-1].split('_')[0].upper()


supported_devices = dict((_key(device), device)
                         for device in Device.__subclasses__())


# TODO: Refactor protocols to use a base class so we can introspect on them.
# Otherwise, consider using a naming convention.
supported_protocols = {
    'DSI': DsiProtocol
}


def find_device(name):
    """Find device by name.

    Parameters
    ----------
        name : str
            name of the device
    Returns
    -------
        Device constructor
    """
    return supported_devices[name]


def protocol_with(name, fs, channels):
    """Find protocol by name and initialize with the given parameters.

    Parameters
    ----------
        name : str
            name of the device
        fs : int
            used to override the default protocol fs value
        channels : list
            used to override the default protocol channels value
    Returns
    -------
        Protocol
    """
    return supported_protocols[name](fs, channels)


def default_protocol(name):
    """Find protocol by name and initialize with default options.

    Parameters
    ----------
        name : str
            name of the device
    Returns
    -------
        Protocol
    """
    return supported_protocols[name]()



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from enum import Enum
from scipy import signal
from scipy.signal import welch
from scipy.integrate import simps
from mne.time_frequency import psd_array_multitaper

from typing import Tuple


class PSD_TYPE(Enum):
    """Power Spectal Density Type.

    Enum used to specify the type of power spectral approximation
        to use. They each have trade-offs in terms of computational
        cost and resolution.
    """
    WELCH = 'Welch'
    MULTITAPER = 'MutliTaper'


def power_spectral_density(
        data: np.ndarray,
        band: Tuple[float, float],
        sampling_rate: float=100.0,
        window_length: float=4.0,
        plot: bool=False,
        method: PSD_TYPE=PSD_TYPE.WELCH,
        relative=False):
    """Power spectral density:

    Many thanks to: https://raphaelvallat.github.io/bandpower.html

    Parameters
    ----------
        data: Numpy Array.
            Time series data in the form of a numpy ndarray used to estimate the
                power spectral density.
        band: tuple.
            frequency band psd to export. Note this must be within the Nyquist frequency
            for your data. Approximated as sampling rate / 2.
        sampling_rate: float
            Sampling rate of the data in # of samples per second.
        window_length: float
            Length in seconds of data window.
        plot: boolean.
            Whether of not to plot the PSD estimated. Helpful for debugging and exploration.
        method: PSD_TYPE
            Type of PSD estimation method to use during calculation.
        relative: boolean
            Whether or not to express the power in a frequency band as a percentage of the
            total power of the signal.

    """
    band = np.asarray(band)
    low, high = band

    # Compute the modified periodogram (Welch)
    if method == PSD_TYPE.WELCH:
        nperseg = window_length * sampling_rate
        freqs, psd = welch(data, sampling_rate, nperseg=nperseg)

    # Compute the modified periodogram (MultiTaper)
    elif method == PSD_TYPE.MULTITAPER:
        psd, freqs = psd_array_multitaper(
            data, sampling_rate, adaptive=True, normalization='full', verbose=False)

    # Find index of band in frequency vector
    idx_band = np.logical_and(freqs >= low, freqs <= high)

    # Frequency resolution
    freq_res = freqs[1] - freqs[0]

    # Integral approximation of the spectrum using parabola (Simpson's rule)
    bp = simps(psd[idx_band], dx=freq_res)

    # Plot the power spectrum
    if plot:
        sns.set(font_scale=1.2, style='white')
        plt.figure(figsize=(8, 4))
        plt.plot(freqs, psd, color='k', lw=2)
        plt.xlabel('Frequency (Hz)')
        plt.ylabel('Power spectral density (V^2 / Hz)')
        plt.ylim([0, psd.max() * 1.1])
        plt.title(f'{method.value}')
        plt.xlim([0, sampling_rate / 2])
        sns.despine()
        plt.show()

    # Whether or not to return PSD as a percentage of total power
    if relative:
        bp /= simps(psd, dx=freq_res)

    return bp


if __name__ == '__main__':
    data = np.loadtxt('bcipy/signal/process/decomposition/resources/data.txt')
    sampling_rate = 100
    band = (0, 100)
    np.arange(data.size) / sampling_rate
    power_spectral_density(data, band, sampling_rate=sampling_rate, method=PSD_TYPE.WELCH, plot=True)



from bcipy.feedback.visual.visual_feedback import VisualFeedback
from psychopy import visual, core

from typing import Tuple


class LevelFeedback(VisualFeedback):
    """Level Feedback.

    A progress bar like feedback to indicate current abilities in a BCI task. This could
        be atteniveness, muscle or eye activity, etc.

    It does not return stimuli timing or allow for parameterized configuration of levels / color gradient
    at this time.
    """

    def __init__(self, display, parameters, clock):
        super().__init__(display, parameters, clock)

        # extract needed parameters
        self.clock = clock
        self.display = display
        self.parameters = parameters
        self.presentation_time = parameters['feedback_flash_time']

        # set stylistic elements
        self.default_message_color = parameters['feedback_message_color']
        self.feedback_indicator_color = parameters['feedback_message_color']
        self.position_x = parameters['feedback_pos_x']
        self.position_y = parameters['feedback_pos_y']
        self.padding = parameters['feedback_padding']
        self.width = parameters['feedback_stim_width']
        self.height = parameters['feedback_stim_height']
        self.line_color = parameters['feedback_line_color']
        self.line_width = parameters['feedback_line_width']
        self.font = parameters['feedback_font']
        self.target_line_width = parameters['feedback_target_line_width']

        # colors needed for each level. This defines the number of levels in the bar indicator.
        self.level_colors = ['red', 'darkorange', 'yellow', 'lightgreen', 'green']

        # array to hold our level feedback stimuli to be presented
        self.stimuli = []

    def administer(self, position=1):
        # construct the bar indicator
        self._construct_bar_indicator()

        # indexing starts at zero but position does not ;)
        position -= 1

        # loop through the levels wanted and construct the stimuli
        for index, color in enumerate(self.level_colors):
            # this is the target indicator level and should stand out of the rest
            if position == index:
                self.stimuli[index].lineColor = self.feedback_indicator_color
                self.stimuli[index].lineWidth = self.target_line_width
            # draw other stimuli elements
            self.stimuli[index].fillColor = color
            self.stimuli[index].draw()

        # flip the display and present for a time
        time = ['visual_feedback', self.clock.getTime()]
        self.display.flip()
        core.wait(self.presentation_time)

        return time

    def _construct_bar_indicator(self):
        """Construct Bar Indicator.

        Helper method for constructing the bar indicator programmatically. It
            resets the position_x and stimuli variables on first call.
        """
        self._reset_stimuli()
        height, width = self._determine_height_and_width()

        # construct the rectangular level shapes and append to stimuli array
        for level in self.level_colors:
            position = self._determine_stimuli_position()
            rect = visual.Rect(
                win=self.display,
                width=width,
                height=height,
                fillColor=self.default_message_color,
                lineColor=self.line_color,
                pos=position,
                lineWidth=self.line_width,
                ori=0.0)
            self.stimuli.append(rect)

    def _determine_stimuli_position(self) -> Tuple[int, int]:
        """Determine Stimuli Position.

        Defines the x and y position of the feedback level stimuli,
            incrementing by the self.padding variable after setting.
            If a vertical bar is desired the position_y should be incremented
            instead of position_x.
        """
        # define the positioning of stimuli and increment the x position
        response = self.position_x, self.position_y
        self.position_x += self.padding
        return response

    def _determine_height_and_width(self) -> Tuple[float, float]:
        """Determine Height and Width.

        Defines the height and width of level stimuli. Currently, a static
            set of variables.
        """
        return self.width, self.height

    def _reset_stimuli(self) -> None:
        """Reset Stimuli.

        Used when redrawing stimuli. Sets the stimuli array and starting
            x positions to their starting values. If using vertical bars
            this should reset the y positions instead.
        """
        # reset stimuli
        self.stimuli = []
        self.position_x = self.parameters['feedback_pos_x']


if __name__ == '__main__':
    from bcipy.helpers.load import load_json_parameters
    from bcipy.display.display_main import init_display_window

    # Load a parameters file
    parameters = load_json_parameters('bcipy/parameters/parameters.json', value_cast=True)
    display = init_display_window(parameters)
    clock = core.Clock()
    feedback = LevelFeedback(display, parameters, clock)

    for index, _ in enumerate(feedback.level_colors):
        # the position argument is based on index starting at 1
        index += 1
        feedback.administer(position=index)



from psychopy import visual
from bcipy.helpers.system_utils import get_system_info


def init_display_window(parameters):
    """
    Init Display Window.

    Function to Initialize main display window
        needed for all later stimuli presentation.

    See Psychopy official documentation for more information and working demos:
        http://www.psychopy.org/api/visual/window.html
    """

    # Check is full_screen mode is set and get necessary values
    if parameters['full_screen']:

        # get relevant info about the system
        info = get_system_info()

        # set window attributes based on resolution
        window_height = info['resolution'][1]
        window_width = info['resolution'][0]

        # set full screen mode to true (removes os dock, explorer etc.)
        full_screen = True

    # otherwise, get user defined window attributes
    else:

        # set window attributes directly from parameters file
        window_height = parameters['window_height']
        window_width = parameters['window_width']

        # make sure full screen is set to false
        full_screen = False

    # Initialize PsychoPy Window for Main Display of Stimuli
    display_window = visual.Window(
        size=[window_width,
              window_height],
        screen=int(parameters['stim_screen']),
        allowGUI=False,
        useFBO=False,
        fullscr=full_screen,
        allowStencil=False,
        monitor='mainMonitor',
        winType='pyglet', units='norm', waitBlanking=False,
        color=parameters['background_color'])

    # Return display window to caller
    return display_window



import logging
import os.path as path
from typing import Callable, List, Optional, Tuple

from psychopy import core, visual

from bcipy.acquisition.marker_writer import NullMarkerWriter, MarkerWriter
from bcipy.helpers.task import SPACE_CHAR
from bcipy.helpers.stimuli import resize_image
from bcipy.helpers.system_utils import get_system_info
from bcipy.helpers.triggers import TriggerCallback, _calibration_trigger


class RSVPDisplay(object):
    """RSVP Display Object for Sequence Presentation.

    Animates a sequence in RSVP. Mode should be determined outside.
    """

    def __init__(
            self,
            window: visual.Window,
            static_clock,
            experiment_clock: core.Clock,
            marker_writer: Optional[MarkerWriter]=None,
            task_color: List[str]=['white'],
            task_font: str='Times',
            task_pos: Tuple[float, float]=(-.8, .9),
            task_height: float=0.2,
            task_text: str='1/100',
            info_color: List[str]=['white'],
            info_text: List[str]=['Information Text'],
            info_font: List[str]=['Times'],
            info_pos=[(.8, .9)],
            info_height=[0.2],
            stim_font='Times',
            stim_pos=(-.8, .9),
            stim_height=0.2,
            stim_sequence: List[str]=['a'] * 10,
            stim_colors: List[str]=['white'] * 10,
            stim_timing: List[float]=[1] * 10,
            is_txt_stim: bool=True,
            static_time: float=.05,
            trigger_type: str='image',
            space_char: SPACE_CHAR=SPACE_CHAR):
        """Initialize RSVP window parameters and objects.

        PARAMETERS:
        ----------
        # Experiment
        window(visual.Window): PsychoPy Window
        static_clock(TODO): no idea
        experiment_clock(core.Clock): Clock used to timestamp experiment
        marker_writer(MarkerWriter): object used to write triggers to
            the daq stream.

        # Task
        task_color(list[string]): Color of the task string. Shares the
            length of the task_text. If of length 1 the entire task
            bar shares the same color.
        task_font(string): Font of task string
        task_pos(tuple): position of task string
        task_height(float): height for task string
        task_text(string): text of the task bar

        # Info
        info_text(list[string]): Text list for information texts
        info_color(list[string]): Color of the information text string
        info_font(list[string]): Font of the information text string
        info_pos(list[tuple]): Position of the information text string
        info_height(list[float]): Height of the information text string

        # Stimuli
        stim_height(float): height of the stimuli object
        stim_pos(tuple): position of stimuli
        stim_font(string): font of the stimuli
        stim_sequence(list[string]): list of elements to flash
        stim_colors(list[string]): list of colors for stimuli
        stim_timing(list[float]): timing for each letter flash
        """
        self.window = window
        self.refresh_rate = window.getActualFrameRate()

        self.logger = logging.getLogger(__name__)

        self.stimuli_sequence = stim_sequence
        self.stimuli_colors = stim_colors
        self.stimuli_timing = stim_timing

        self.is_txt_stim = is_txt_stim
        self.staticPeriod = static_clock
        self.static_time = static_time
        self.experiment_clock = experiment_clock
        self.timing_clock = core.Clock()

        # Used to handle writing the marker stimulus
        self.marker_writer = marker_writer or NullMarkerWriter()

        # Length of the stimuli (number of flashes)
        self.stim_length = len(stim_sequence)

        # Informational Parameters
        self.info_text = info_text

        # Stim parameters
        self.stimuli_font = stim_font
        self.stimuli_height = stim_height
        self.stimuli_pos = stim_pos

        # Trigger Items
        self.first_run = True
        self.trigger_type = trigger_type
        self.trigger_callback = TriggerCallback()

        # Callback used on presentation of first stimulus.
        self.first_stim_callback = lambda _sti: None
        self.size_list_sti = []

        self.space_char = space_char

        self.task = visual.TextStim(win=self.window, color=task_color[0],
                                    height=task_height,
                                    text=task_text,
                                    font=task_font, pos=task_pos,
                                    wrapWidth=None, colorSpace='rgb',
                                    opacity=1, depth=-6.0)

        # Create multiple text objects based on input
        self.text = []
        for idx in range(len(self.info_text)):
            self.text.append(visual.TextStim(
                win=self.window,
                color=info_color[idx],
                height=info_height[idx],
                text=self.info_text[idx],
                font=info_font[idx],
                pos=info_pos[idx],
                wrapWidth=None, colorSpace='rgb',
                opacity=1, depth=-6.0))

        # Create Stimuli Object
        if self.is_txt_stim:
            self.sti = visual.TextStim(
                win=self.window,
                color='white',
                height=self.stimuli_height,
                text='+',
                font=self.stimuli_font,
                pos=self.stimuli_pos,
                wrapWidth=None, colorSpace='rgb',
                opacity=1, depth=-6.0)
        else:
            self.sti = visual.ImageStim(
                win=self.window,
                image=None,
                mask=None,
                pos=self.stimuli_pos,
                ori=0.0)

    def draw_static(self):
        """Draw static elements in a stimulus."""
        self.task.draw()
        for idx in range(len(self.text)):
            self.text[idx].draw()

    def schedule_to(self, ele_list=[], time_list=[], color_list=[]):
        """Schedule stimuli elements (works as a buffer).

        Args:
                ele_list(list[string]): list of elements of stimuli
                time_list(list[float]): list of timings of stimuli
                color_list(list[string]): colors of elements of stimuli
        """
        self.stimuli_sequence = ele_list
        self.stimuli_timing = time_list
        self.stimuli_colors = color_list

    def update_task(self, text: str, color_list: List[str], pos: Tuple[float]):
        """Update Task Object.

        PARAMETERS:
        -----------
        text: text for task
        color_list: list of the colors for each char
        pos: position of task
        """
        self.task.text = text
        self.task.color = color_list[0]
        self.task.pos = pos

    def do_sequence(self):
        """Do Sequence.

        Animates a sequence of flashing letters to achieve RSVP.
        """

        # init an array for timing information
        timing = []

        if self.first_run:
            # play a sequence start sound to help orient triggers
            first_stim_timing = _calibration_trigger(
                self.experiment_clock,
                trigger_type=self.trigger_type, display=self.window,
                on_trigger=self.marker_writer.push_marker)

            timing.append(first_stim_timing)

            self.first_stim_time = first_stim_timing[-1]
            self.first_run = False

        # do the sequence
        for idx in range(len(self.stimuli_sequence)):

            # set a static period to do all our stim setting.
            #   will warn if ISI value is violated.
            self.staticPeriod.start(self.static_time)

            # turn ms timing into frames! Much more accurate!
            self.time_to_present = int(self.stimuli_timing[idx] * self.refresh_rate)

            # check if stimulus needs to use a non-default size
            if self.size_list_sti:
                this_stimuli_size = self.size_list_sti[idx]
            else:
                this_stimuli_size = self.stimuli_height

            # Set the Stimuli attrs
            if self.stimuli_sequence[idx].endswith('.png'):
                self.sti = self.create_stimulus(mode='image', height_int=this_stimuli_size)
                self.sti.image = self.stimuli_sequence[idx]
                self.sti.size = resize_image(self.sti.image, self.sti.win.size, this_stimuli_size)
                sti_label = path.splitext(
                    path.basename(self.stimuli_sequence[idx]))[0]
            else:
                # text stimulus
                self.sti = self.create_stimulus(mode='text', height_int=this_stimuli_size)
                txt = self.stimuli_sequence[idx]
                # customize presentation of space char.
                self.sti.text = txt if txt != SPACE_CHAR else self.space_char
                self.sti.color = self.stimuli_colors[idx]
                sti_label = txt

                # test whether the word will be too big for the screen
                text_width = self.sti.boundingBox[0]
                if text_width > self.window.size[0]:
                    info = get_system_info()
                    text_height = self.sti.boundingBox[1]
                    # If we are in full-screen, text size in Psychopy norm units
                    # is monitor width/monitor height
                    if self.window.size[0] == info['RESOLUTION'][0]:
                        new_text_width = info['RESOLUTION'][0] / info['RESOLUTION'][1]
                    else:
                        # If not, text width is calculated relative to both
                        # monitor size and window size
                        new_text_width = (
                            self.window.size[1] / info['RESOLUTION'][1]) * (
                                info['RESOLUTION'][0] / info['RESOLUTION'][1])
                    new_text_height = (text_height * new_text_width) / text_width
                    self.sti.height = new_text_height

            # End static period
            self.staticPeriod.complete()

            # Reset the timing clock to start presenting
            self.window.callOnFlip(self.trigger_callback.callback, self.experiment_clock, sti_label)
            self.window.callOnFlip(self.marker_writer.push_marker, sti_label)

            if idx == 0 and callable(self.first_stim_callback):
                self.first_stim_callback(self.sti)

            # Draw stimulus for n frames
            for _n_frames in range(self.time_to_present):
                self.sti.draw()
                self.draw_static()
                self.window.flip()

            # append timing information
            if self.is_txt_stim:
                timing.append(self.trigger_callback.timing)
            else:
                timing.append(self.trigger_callback.timing)

            self.trigger_callback.reset()

        # draw in static and flip once more
        self.draw_static()
        self.window.flip()

        return timing

    def update_task_state(self, text: str, color_list: List[str]) -> None:
        """Update task state.

        Removes letters or appends to the right.
        Args:
                text(string): new text for task state
                color_list(list[string]): list of colors for each
        """
        task_state_text = visual.TextStim(win=self.window, font=self.task.font, text=text)
        x_task_position = task_state_text.boundingBox[0] / self.window.size[0] - 1
        task_pos = (x_task_position, 1 - self.task.height)

        self.update_task(text=text, color_list=color_list, pos=task_pos)

    def wait_screen(self, message, color):
        """Wait Screen.

        Args:
            message(string): message to be displayed while waiting
        """

        # Construct the wait message
        wait_message = visual.TextStim(win=self.window, font=self.stimuli_font,
                                       text=message,
                                       height=.1,
                                       color=color,
                                       pos=(0, -.5),
                                       wrapWidth=2,
                                       colorSpace='rgb',
                                       opacity=1, depth=-6.0)

        # Try adding our BCI logo. Pass if not found.
        try:
            wait_logo = visual.ImageStim(
                self.window,
                image='bcipy/static/images/gui_images/bci_cas_logo.png',
                pos=(0, .5),
                mask=None,
                ori=0.0)
            wait_logo.size = resize_image(
                'bcipy/static/images/gui_images/bci_cas_logo.png',
                self.window.size, 1)
            wait_logo.draw()

        except Exception:
            self.logger.debug('Cannot load logo image')
            pass

        # Draw and flip the screen.
        wait_message.draw()
        self.window.flip()

    def create_stimulus(self, height_int: int, mode: str='text'):
        """Create Stimulus.

        Returns a TextStim or ImageStim object.
            Args:
            height_int: The height of the stimulus
            mode: "text" or "image", determines which to return
        """
        if mode == 'text':
            return visual.TextStim(
                win=self.window,
                color='white',
                height=height_int,
                text='+',
                font=self.stimuli_font,
                pos=self.stimuli_pos,
                wrapWidth=None,
                colorSpace='rgb',
                opacity=1,
                depth=-6.0)
        if mode == 'image':
            return visual.ImageStim(
                win=self.window,
                image=None,
                mask=None,
                units='',
                pos=self.stimuli_pos,
                size=(height_int, height_int),
                ori=0.0)



from enum import Enum
from bcipy.language_model.lm_server import LmServerConfig
from bcipy.language_model import oclm_language_model
from bcipy.language_model import prelm_language_model
from bcipy.helpers.system_utils import dot


class LmType(Enum):
    """Enum of the registered language model types. The types are associated
    with constructors for creating the model.
    Ex.
    >>> LmType.PRELM.model()
    """
    PRELM = prelm_language_model.LangModel
    OCLM = oclm_language_model.LangModel

    # pylint: disable=unused-argument,protected-access
    def __new__(cls, *args, **kwds):
        """Autoincrements the value of each item added to the enum."""
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, model):
        self.model = model


def LangModel(lmtype: LmType, logfile: str = "log", port: int = None):
    """Creates a new Language Model given the LmType."""

    assert lmtype, "Language Model type is required"
    model = lmtype.model
    config = model.DEFAULT_CONFIG
    if port:
        config.port = port
    return model(config, logfile)



from bcipy.feedback.feedback import Feedback
from psychopy import visual, core
from bcipy.helpers.stimuli import resize_image
from enum import Enum


class FeedbackType(Enum):
    TEXT = 'TEXT'
    IMAGE = 'IMAGE'
    SHAPE = 'SHAPE'


class VisualFeedback(Feedback):
    """Visual Feedback."""

    def __init__(self, display, parameters, clock):

        # Register Feedback Type
        self.feedback_type = 'Visual Feedback'

        super(VisualFeedback, self).__init__(self.feedback_type)

        # Display Window
        self.display = display

        # Parameters
        self.parameters = parameters
        self.font_stim = self.parameters['feedback_font']
        self.height_stim = self.parameters['feedback_stim_height']
        self.width_stim = self.parameters['feedback_stim_width']

        pos_x = self.parameters['feedback_pos_x']
        pos_y = self.parameters['feedback_pos_y']

        self.pos_stim = (pos_x, pos_y)

        self.feedback_length = self.parameters['feedback_flash_time']

        # Clock
        self.clock = clock

        self.message_color = self.parameters['feedback_message_color']
        self.feedback_line_width = self.parameters['feedback_line_width']

    def administer(
            self,
            stimulus,
            pos=None,
            line_color='blue',
            fill_color='blue',
            message=None,
            compare_assertion=None,
            stimuli_type=FeedbackType.TEXT):
        """Administer.

        Administer visual feedback. Timing information from parameters,
            current feedback given by stimulus, if assertion type feedback
            wanted, it's added as an optional argument.
        """
        timing = []

        if message:
            message = self._construct_message(message)
            message.draw()

        if compare_assertion:
            (stim,
             assert_stim) = self._construct_assertion_stimuli(
                stimulus,
                compare_assertion,
                line_color,
                fill_color,
                stimuli_type)

            assert_stim.draw()
        else:
            stim = self._construct_stimulus(
                stimulus,
                self.pos_stim,
                line_color,
                fill_color,
                stimuli_type)

        self._show_stimuli(stim)
        time = ['visual_feedback', self.clock.getTime()]

        core.wait(self.feedback_length)
        timing.append(time)

        return timing

    def _show_stimuli(self, stimulus):
        stimulus.draw()
        self.display.flip()

    def _construct_stimulus(self, stimulus, pos, line_color, fill_color, stimuli_type):
        if stimuli_type == FeedbackType.IMAGE:
            image_stim = visual.ImageStim(
                win=self.display,
                image=stimulus,
                mask=None,
                pos=pos,
                ori=0.0)
            image_stim.size = resize_image(stimulus, self.display.size, self.height_stim)
            return image_stim
        if stimuli_type == FeedbackType.TEXT:
            return visual.TextStim(
                win=self.display,
                font=self.font_stim,
                text=stimulus,
                height=self.height_stim,
                pos=pos,
                color=fill_color)
        if stimuli_type == FeedbackType.SHAPE:
            return visual.Rect(
                fillColor=fill_color, win=self.display,
                width=self.width_stim,
                height=self.height_stim,
                lineColor=line_color,
                pos=(self.pos_stim),
                lineWidth=self.feedback_line_width,
                ori=0.0)

    def _construct_assertion_stimuli(
            self,
            stimulus,
            assertion,
            line_color,
            fill_color,
            stimuli_type):
        stimulus = self._construct_stimulus(stimulus, (-.3, 0), line_color, fill_color, stimuli_type)
        assertion = self._construct_stimulus(assertion, (.3, 0), line_color, fill_color, stimuli_type)

        return stimulus, assertion

    def _construct_message(self, message):
        return visual.TextStim(win=self.display, font=self.font_stim,
                               text=message,
                               height=0.3,
                               pos=(-.3, .5), color=self.message_color)


if __name__ == "__main__":
    import argparse
    from bcipy.helpers.load import load_json_parameters
    from bcipy.display.display_main import init_display_window

    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--parameters',
                        default='bcipy/parameters/parameters.json',
                        help='Parameter location. Must be in parameters directory. \
                          Pass as bcipy/parameters/parameters.json')

    args = parser.parse_args()

    # Load a parameters file
    parameters = load_json_parameters(args.parameters, value_cast=True)
    display = init_display_window(parameters)
    clock = core.Clock()
    # Start Visual Feedback
    visual_feedback = VisualFeedback(
        display=display, parameters=parameters, clock=clock)
    stimulus = 'null'
    timing = visual_feedback.administer(
        stimulus, line_color='blue', fill_color='blue', stimuli_type=FeedbackType.SHAPE)
    print(timing)
    print(visual_feedback._type())



"""Data server that streams EEG data over a LabStreamingLayer StreamOutlet
using pylsl."""
import logging
import random
from queue import Queue, Empty
from pylsl import StreamInfo, StreamOutlet
from bcipy.acquisition.datastream.producer import Producer
from bcipy.acquisition.util import StoppableThread

log = logging.getLogger(__name__)

# pylint: disable=too-many-arguments


class LslDataServer(StoppableThread):
    """Data server that streams EEG data over a LabStreamingLayer StreamOutlet
    using pylsl. See https://github.com/sccn/labstreaminglayer/wiki.

    In parameters.json, if the fake_data parameter is set to true and the
    device is set to LSL, this server will be used to mock data. Alternatively,
    fake_data can be set to false and this module can be run standalone in its
    own python instance.

    Parameters
    ----------
        params : dict(channels: list(str), hz(int))
            parameters used to configure the server. Should have at least
            a list of channels and the sample frequency.
        generator : Object (see generator.py for options)
            used to generate the data to be served. Ex. random_data.
        include_meta: bool, optional
            if True, writes metadata to the outlet stream.
        add_markers: bool, optional
            if True, creates a the marker channel and streams data to this
            channel at a fixed frequency.
    """

    def __init__(self, params, generator, include_meta=True,
                 add_markers=False, name='TestStream'):
        super(LslDataServer, self).__init__()

        channels = params['channels']
        self.sample_hz = int(params['hz'])
        assert channels, "Channels must be provided as a parameter"

        stream_name = params.get('name', name)
        self.generator = generator

        log.debug("Starting server with params: %s", str(params))
        info = StreamInfo(stream_name,
                          "EEG", len(channels),
                          self.sample_hz,
                          'float32',
                          "uid12345")

        if include_meta:
            meta_channels = info.desc().append_child('channels')
            for channel in channels:
                meta_channels.append_child('channel') \
                    .append_child_value('label', channel) \
                    .append_child_value('unit', 'microvolts') \
                    .append_child_value('type', 'EEG')

        self.outlet = StreamOutlet(info)

        self.add_markers = add_markers
        if add_markers:
            # Marker stream (C++ source below):
            # lsl::stream_info marker_info("gUSBamp-"+deviceNumber+"Markers",
            # "Markers",1,0,lsl::cf_string,
            # "gUSBamp_" + boost::lexical_cast<std::string>(deviceNumber) +
            # "_" + boost::lexical_cast<std::string>(serialNumber) +
            # "_markers");
            log.debug("Creating marker stream")
            markers_info = StreamInfo("TestStream Markers",
                                      "Markers", 1, 0, 'string',
                                      "uid12345_markers")
            self.markers_outlet = StreamOutlet(markers_info)
        self.started = False

    def stop(self):
        """Stop the thread and cleanup resources."""

        log.debug("[*] Stopping data server")
        super(LslDataServer, self).stop()

        # Allows pylsl to cleanup; The outlet will no longer be discoverable
        # after destruction and all connected inlets will stop delivering data.
        del self.outlet
        self.outlet = None

        if self.add_markers:
            del self.markers_outlet
            self.markers_outlet = None

    def run(self):
        """Main loop of the thread. Continuously streams data to the stream
        outlet at a rate consistent with the sample frequency. May also
        output markers at a different interval."""

        sample_counter = 0
        self.started = True

        data_queue = Queue()
        with Producer(data_queue, generator=self.generator,
                      freq=1 / self.sample_hz):
            while self.running():
                sample_counter += 1
                try:
                    sample = data_queue.get(True, 2)
                    self.outlet.push_sample(sample)
                    if self.add_markers and sample_counter % 1000 == 0:
                        self.markers_outlet.push_sample(
                            [str(random.randint(1, 100))])
                except (Empty, AttributeError):
                    # outlet.push_sample(sample) may cause an error after
                    # the server has been stopped since the attribute is
                    # deleted in another thread.
                    break

        log.debug("[*] No longer pushing data")


def _settings(filename):
    """Read the daq settings from the given data file"""

    with open(filename, 'r') as datafile:
        daq_type = datafile.readline().strip().split(',')[1]
        sample_hz = int(datafile.readline().strip().split(',')[1])
        channels = datafile.readline().strip().split(',')
        return (daq_type, sample_hz, channels)


def main():
    """Initialize and start the server."""
    import time
    import argparse

    from bcipy.acquisition.datastream.generator import file_data, random_data

    default_channels = ['ch' + str(i + 1) for i in range(16)]

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--filename', default=None,
                        help="file containing data to be streamed; "
                        "if missing, random data will be served.")
    parser.add_argument('-c', '--channels',
                        default=','.join(default_channels),
                        help='comma-delimited list')
    parser.add_argument('-s', '--sample_rate', default='256',
                        help='sample rate in hz')

    parser.add_argument('-m', '--markers', action="store_true", default=False)
    parser.add_argument('-n', '--name', default='LSL')
    args = parser.parse_args()

    params = {'channels': args.channels.split(','),
              'hz': int(args.sample_rate)}

    # Generate data from the file if provided, otherwise random data.
    generator = file_data(filename=args.filename) if args.filename \
        else random_data(channel_count=len(params['channels']))

    markers = True if args.markers else False
    try:
        server = LslDataServer(params=params, generator=generator,
                               add_markers=markers, name=args.name)

        log.debug("New server created")
        server.start()
        log.debug("Server started")
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("Keyboard Interrupt")
        server.stop()


if __name__ == '__main__':
    # Run this as: python -m bcipy.acquisition.datastream.lsl_server
    main()



import unittest
from io import StringIO
from typing import List, Tuple
import random
from bcipy.helpers.triggers import NONE_VALUE, LslCopyPhraseLabeller, \
    extract_from_copy_phrase, extract_from_calibration, \
    write_trigger_file_from_lsl_calibration


def sample_raw_data(trigger_seq: List[Tuple[str, str]] =[],
                    first_trg_time: int=100,
                    trigger_interval: int=10) -> Tuple[str, List[float]]:
    """Helper function for creating mock data that looks like the raw_data.csv
    output. Adds trigger data to the TRG column at the specified interval.

    Parameters:
    -----------
        trigger_seq: list of tuples with the stimulus, targetness.
        first_trg_time: first time in the data where a trigger should appear.
        trigger_interval: set interval at which subsequent triggers should be
                    displayed
    Returns:
    --------
        content: str, trigger_times: list(float)
    """
    # Set up the trigger times
    trigger_times = []
    triggers_by_time = {}
    for i in range(len(trigger_seq)):
        ts = first_trg_time + (i * trigger_interval)
        trigger_times.append(ts)
        trg_val = trigger_seq[i][0]
        triggers_by_time[ts] = trg_val

    # Mock the raw_data file
    sep = '\r\n'
    meta = sep.join(['daq_type,TestStream', 'sample_rate,300'])
    header = 'timestamp,c1,c2,c3,TRG'

    data = []
    for i in range(1000):
        timestamp = i + 10.0
        channel_data = [str(random.uniform(-1000, 1000)) for _ in range(3)]
        trg = triggers_by_time.get(timestamp, NONE_VALUE)
        data.append(','.join([str(timestamp), *channel_data, trg]))

    content = sep.join([meta, header, *data])
    return content, trigger_times


class TestTriggers(unittest.TestCase):
    """This is Test Case for Triggers."""

    def test_triggers_for_calibration(self):
        return

    def test_copy_phrase_labeller(self):
        copy_phrase = 'HI'
        typed = 'HI'

        labeller = LslCopyPhraseLabeller(copy_phrase, typed)
        self.assertEqual('calib',
                         labeller.label("['calibration_trigger', "
                                        "2.30196808103]"))
        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('nontarget', labeller.label('A'))
        self.assertEqual('nontarget', labeller.label('B'))
        self.assertEqual('target', labeller.label('H'))
        self.assertEqual('nontarget', labeller.label('I'))

        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('target', labeller.label('I'))
        self.assertEqual('nontarget', labeller.label('A'))
        self.assertEqual('nontarget', labeller.label('H'))
        self.assertEqual('nontarget', labeller.label('O'))

    def test_copy_phrase_labeller_correction(self):
        copy_phrase = 'HI'
        typed = 'HA<I'

        labeller = LslCopyPhraseLabeller(copy_phrase, typed)
        self.assertEqual('calib',
                         labeller.label("['calibration_trigger', "
                                        "2.30196808103]"))
        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('nontarget', labeller.label('B'))
        self.assertEqual('target', labeller.label('H'))
        self.assertEqual('nontarget', labeller.label('I'))

        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('target', labeller.label('I'))
        self.assertEqual('nontarget', labeller.label('A'))
        self.assertEqual('nontarget', labeller.label('O'))

        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('nontarget', labeller.label('I'))
        self.assertEqual('nontarget', labeller.label('A'))
        self.assertEqual('target', labeller.label('<'))

        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('target', labeller.label('I'))
        self.assertEqual('nontarget', labeller.label('A'))
        self.assertEqual('nontarget', labeller.label('O'))

    def test_copy_phrase_labeller_correction_double_letters(self):
        copy_phrase = 'HELLO'
        typed = 'HELP<LO'

        labeller = LslCopyPhraseLabeller(copy_phrase, typed)
        self.assertEqual('calib',
                         labeller.label("['calibration_trigger', "
                                        "2.30196808103]"))
        # H
        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('target', labeller.label('H'))
        self.assertEqual('nontarget', labeller.label('I'))

        # E
        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('target', labeller.label('E'))
        self.assertEqual('nontarget', labeller.label('O'))

        # L
        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('nontarget', labeller.label('O'))
        self.assertEqual('target', labeller.label('L'))

        # L
        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('target', labeller.label('L'))
        self.assertEqual('nontarget', labeller.label('O'))

        # <
        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('nontarget', labeller.label('O'))
        self.assertEqual('target', labeller.label('<'))

        # L
        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('target', labeller.label('L'))
        self.assertEqual('nontarget', labeller.label('A'))

        # O
        self.assertEqual('fixation', labeller.label('+'))
        self.assertEqual('nontarget', labeller.label('A'))
        self.assertEqual('target', labeller.label('O'))

    def test_extract_from_copy_phrase(self):
        trigger_seq = [
            ("['calibration_trigger', 2.3019680810393766]", 'calib'),
            ('+', 'fixation'), ('I', 'nontarget'), ('H', 'nontarget'),
            ('C', 'nontarget'), ('G', 'nontarget'), ('D', 'nontarget'),
            ('F', 'nontarget'), ('_', 'nontarget'), ('E', 'nontarget'),
            ('<', 'nontarget'), ('B', 'nontarget'), ('+', 'fixation'),
            ('E', 'nontarget'), ('B', 'nontarget'), ('H', 'nontarget'),
            ('_', 'nontarget'), ('C', 'nontarget'), ('I', 'nontarget'),
            ('G', 'nontarget'), ('<', 'nontarget'), ('F', 'nontarget'),
            ('D', 'nontarget'), ('+', 'fixation'), ('G', 'nontarget'),
            ('_', 'nontarget'), ('B', 'nontarget'), ('F', 'nontarget'),
            ('I', 'nontarget'), ('C', 'nontarget'), ('<', 'nontarget'),
            ('E', 'nontarget'), ('D', 'nontarget'), ('H', 'nontarget')]

        phrase = 'HELLO'
        # TODO: why does the copy phrase task starts in the middle of the
        # phrase?
        start_index = int(len(phrase) / 2)
        copy_text = phrase[start_index:]  # 'LLO'
        content, trigger_times = sample_raw_data(trigger_seq)
        extracted = extract_from_copy_phrase(StringIO(content),
                                             copy_text=copy_text,
                                             typed_text=copy_text)

        # Assertions
        self.assertEqual(len(trigger_seq), len(extracted))
        for seq_i in range(len(trigger_seq)):
            extracted_val, extracted_targetness, stamp = extracted[seq_i]
            expected_trg, targetness = trigger_seq[seq_i]
            if 'calibration' in expected_trg:
                expected_trg = 'calibration_trigger'
            self.assertEqual(expected_trg, extracted_val)
            self.assertEqual(targetness, extracted_targetness)
            self.assertEqual(trigger_times[seq_i], float(stamp))

    def test_extract_from_calibration(self):
        trigger_seq = [
            ('"[\'calibration_trigger\', 2.039073024992831]"', 'calib'),
            ('J', 'first_pres_target'), ('+', 'fixation'), ('P', 'nontarget'),
            ('R', 'nontarget'), ('E', 'nontarget'), ('K', 'nontarget'),
            ('A', 'nontarget'), ('J', 'target'), ('X', 'nontarget'),
            ('F', 'nontarget'), ('<', 'nontarget'), ('S', 'nontarget'),
            ('E', 'first_pres_target'), ('+', 'fixation'), ('M', 'nontarget'),
            ('T', 'nontarget'), ('H', 'nontarget'), ('W', 'nontarget'),
            ('Y', 'nontarget'), ('V', 'nontarget'), ('E', 'target'),
            ('L', 'nontarget'), ('_', 'nontarget'), ('J', 'nontarget')]

        # Mock the raw_data file
        content, trigger_times = sample_raw_data(trigger_seq)
        extracted = extract_from_calibration(StringIO(content),
                                             seq_len=10)

        # Assertions
        self.assertEqual(len(trigger_seq), len(extracted))
        for seq_i in range(len(trigger_seq)):
            extracted_val, extracted_targetness, stamp = extracted[seq_i]
            expected_trg, targetness = trigger_seq[seq_i]
            if 'calibration' in expected_trg:
                expected_trg = 'calibration_trigger'
            self.assertEqual(expected_trg, extracted_val)
            self.assertEqual(targetness, extracted_targetness)
            self.assertEqual(trigger_times[seq_i], float(stamp))

    def test_writing_trigger_file(self):
        trigger_seq = [
            ('"[\'calibration_trigger\', 2.039073024992831]"', 'calib'),
            ('J', 'first_pres_target'), ('+', 'fixation'), ('P', 'nontarget'),
            ('R', 'nontarget'), ('E', 'nontarget'), ('K', 'nontarget'),
            ('A', 'nontarget'), ('J', 'target'), ('X', 'nontarget'),
            ('F', 'nontarget'), ('<', 'nontarget'), ('S', 'nontarget'),
            ('E', 'first_pres_target'), ('+', 'fixation'), ('M', 'nontarget'),
            ('T', 'nontarget'), ('H', 'nontarget'), ('W', 'nontarget'),
            ('Y', 'nontarget'), ('V', 'nontarget'), ('E', 'target'),
            ('L', 'nontarget'), ('_', 'nontarget'), ('J', 'nontarget')]

        # Mock the raw_data file
        raw_data, trigger_times = sample_raw_data(trigger_seq)
        output = StringIO()
        write_trigger_file_from_lsl_calibration(StringIO(raw_data), output,
                                                seq_len=10)

        written_contents = output.getvalue()
        lines = written_contents.split("\n")

        for i in range(len(lines) - 1):
            written_val, written_targetness, written_stamp = lines[i].split()
            expected_trg, targetness = trigger_seq[i]
            if 'calibration' in expected_trg:
                expected_trg = 'calibration_trigger'
            self.assertEqual(expected_trg, written_val)
            self.assertEqual(targetness, written_targetness)
            self.assertEqual(trigger_times[i], float(written_stamp))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""Defines the Record namedtuple"""
from collections import namedtuple

Record = namedtuple('Record', ['data', 'timestamp', 'rownum'])
Record.__doc__ = """Domain object used for storing data and timestamp
information, where data is a single reading from a device and is a list
of channel information (float)."""



"""Utility functions for protocols."""


def receive(socket, msglen, chunksize=2048):
    """Receive an entire message from a socket, which may be chunked.

    Parameters
    ----------
        socket : socket.socket
            object from which to receive data
        msglen : int
            length of the entire message
        chunksize : int ; optional
            messages will be received in chunksize lengths and joined together.

    Return
    ------
        str - entire message.
    """

    chunks = []
    bytes_received = 0
    while bytes_received < msglen:
        recv_len = min(msglen - bytes_received, chunksize)
        chunk = socket.recv(recv_len)
        if chunk == '':
            raise RuntimeError("socket connection broken")
        chunks.append(chunk)
        bytes_received = bytes_received + len(chunk)
    return b''.join(chunks)



"""Task Registry ; used to provide task options to the GUI and command line
tools. User defined tasks can be added to the Registry."""


# NOTE:
# In the future we may want to consider dynamically retrieving all subclasses
# of Task and use these to populate a registry. We could also provide
# functionality for bcipy users to define their own tasks and register them so
# they would appear as options in the GUI.
#
# However, this approach is currently problematic for the GUI interface. Due
# to the tight coupling of the display code with the Tasks, importing any of
# the Task subclasses causes pygame (a psychopy dependency) to create a GUI,
# which seems to prevent our other GUI code from working.

from enum import Enum
from itertools import groupby
from typing import Dict, List


class ExperimentType(Enum):
    """Enum of the registered experiment types (Tasks), along with the label
    used for display in the GUI and command line tools. Values are looked up
    by their (1-based) index.

    Examples:
    >>> ExperimentType(1)
    <ExperimentType.RSVP_CALIBRATION: 1>

    >>> ExperimentType(1).label
    'Calibration'
    """

    RSVP_CALIBRATION = 'Calibration'
    RSVP_COPY_PHRASE = 'Copy Phrase'
    RSVP_ICON_TO_ICON = 'Icon to Icon'
    RSVP_ICON_TO_WORD = 'Icon to Word'
    RSVP_ALERT_TONE_CALIBRATION = 'Alert Tone'
    RSVP_INTER_SEQUENCE_FEEDBACK_CALIBRATION = 'Feed. Calib.'

    # pylint: disable=unused-argument,protected-access
    def __new__(cls, *args, **kwds):
        """Autoincrements the value of each item added to the enum."""
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, label):
        self.label = label

    @classmethod
    def by_mode(cls) -> Dict[str, List['ExperimentType']]:
        """ExperimentTypes by mode (ex. 'RSVP')"""

        def prefix(item, sep='_') -> str:
            name = item.name
            return name[0:name.index(sep)] if sep in name else ''

        mode_map = {}
        for key, group in groupby(cls, prefix):
            mode_map[key] = [task for task in group]
        return mode_map

    @classmethod
    def calibration_tasks(cls) -> List['ExperimentType']:
        return [task for task in cls
                if task.name.endswith('CALIBRATION') and 'COPY_PHRASE'
                not in task.name]
